# """
# Implementação de um decisor (amostrador e quantizador) para recepção.

# Autor: Arthur Cadore
# Data: 15-08-2025
# """

import numpy as np
from .plotter import save_figure, create_figure, SampledSignalPlot

class Sampler:
    def __init__(self, fs=128_000, Rb=400, t=None, delay=0.08):
        r"""
        Inicializa o decisor, utilizado para amostragem e quantização no receptor.

        Args: 
            fs (int): Frequência de amostragem.
            delay (float): Delay de amostragem, em segundos.
            Rb (int): Taxa de bits.
            t (numpy.ndarray): Vetor de tempo.

        Example: 
            ![pageplot](assets/example_sampler_time.svg) 
        """
        self.fs = fs
        self.Rb = Rb
        self.sps = int(self.fs / self.Rb)
        self.delay = int(round(delay * self.fs))

        if t is not None:
            self.indexes = self.calc_indexes(t)

    def update_sampler(self, delay, t):
        self.delay = int(round(delay * self.fs))
        self.indexes = self.calc_indexes(t)
    
    def calc_indexes(self, t):
        r"""
        Calcula os índices de amostragem $I[n]$ com base no vetor de tempo $t$. O vetor de índices de amostragem $I[n]$ é dado pela expressão abaixo. 

        $$
        \begin{align}
        I[n] = \tau + n \cdot \left( \frac{f_s}{R_b}\right) \text{ , onde: } \quad I[n] < \text{len}(t)
        \end{align}
        $$

        Sendo:
            - $\tau$: Delay inicial de amostragem.
            - $f_s$: Frequência de amostragem.
            - $R_b$: Taxa de bits.
            - $n$: Índice da amostra.
            - $\text{len}(t)$: Comprimento do vetor de tempo.

        Args:
            t (numpy.ndarray): Vetor de tempo.

        Returns:
            indexes (numpy.ndarray): Vetor de índices de amostragem $I[n]$.
        """
        indexes = np.arange(self.delay, len(t), self.sps)
        indexes = indexes[indexes < len(t)]
        return indexes
    
    def sample(self, signal):
        r"""
        Amostra o sinal $s(t)$ com base nos índices de amostragem $I[n]$.

        $$
            s(t) \rightarrow  s([I[n]) \rightarrow s[n]
        $$

        Sendo:
            - $s(t)$: Sinal de entrada $s(t)$.
            - $s[n]$ Sinal amostrado $s[n]$.
            - $I[n]$ Índices de amostragem $I[n]$.

        Args:
            signal (numpy.ndarray): Sinal de entrada $s(t)$ a ser amostrado.

        Returns:
            sampled_signal (numpy.ndarray): Sinal amostrado $s[n]$.
        """
        sampled_signal = signal[self.indexes]
        return sampled_signal

    def quantize(self, signal):
        r"""
        Quantiza o sinal $s[n]$ em valores discretos. O processo de quantização é dado pela expressão abaixo.

        $$
        \begin{align}
        s'[n] = \begin{cases}
            +1 & \text{se } s[n] \geq 0 \\
            -1 & \text{se } s[n] < 0
        \end{cases}
        \end{align}
        $$

        Sendo:
            - $s[n]$ Simbolos amostrados $s[n]$.
            - $s'[n]$ Símbolos quantizados $s'[n]$.

        Args:
            signal (numpy.ndarray): Sinal de entrada $s[n]$.

        Returns:
            symbols (numpy.ndarray): Símbolos quantizados $s'[n]$.
        """
        symbols = []
        for i in range(len(signal)):
            if signal[i] >= 0:
                symbols.append(+1)
            else:
                symbols.append(-1)
        return symbols

if __name__ == "__main__":

    fs = 128_000
    Rb = 2000
    t = np.arange(100000) / fs
    signal = np.cos(2 * np.pi * 400 * t) + np.cos(2 * np.pi * 100 * t)
    signal2 = np.sin(2 * np.pi * 400 * t) + np.sin(2 * np.pi * 1000 * t)

    sampler = Sampler(fs=fs, Rb=Rb, t=t)
    sampled_signal = sampler.sample(signal)
    sampled_time = sampler.sample(t)

    sampler2 = Sampler(fs=fs, Rb=Rb, t=t)
    sampled_signal2 = sampler2.sample(signal2)
    sampled_time2 = sampler2.sample(t)

    symbols = sampler.quantize(sampled_signal)
    symbols2 = sampler2.quantize(sampled_signal2)
    print(symbols[:20], "...")
    print(symbols2[:20], "...")

    fig_sampler, grid_sampler = create_figure(2, 1, figsize=(16, 9))

    SampledSignalPlot(
        fig_sampler, grid_sampler, (0, 0),
        t,
        signal,
        sampled_time,
        sampled_signal,
        colors='red',
        label_signal="Sinal original", 
        label_samples="Amostras", 
        xlim=(40, 100), 
        title="Sinal $Cos(t)$ amostrado"
    ).plot()

    SampledSignalPlot(
        fig_sampler, grid_sampler, (1, 0),
        t,
        signal2,
        sampled_time2,
        sampled_signal2,
        colors='navy',
        label_signal="Sinal original", 
        label_samples="Amostras", 
        xlim=(40, 100), 
        title="Sinal $Sin(t)$ amostrado"
    ).plot()

    fig_sampler.tight_layout()
    save_figure(fig_sampler, "example_sampler_time.pdf")


# """
# Implementação do embaralhador e desembaralhador compatível com o padrão PPT-A3.

# Referência:
#     AS3-SP-516-274-CNES (3.1.4.5)

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from .plotter import save_figure, create_figure, BitsPlot

class Scrambler:
    def __init__(self):
        r"""
        Inicializa o embaralhador no padrão ARGOS-3.

        Example: 
            ![pageplot](assets/example_scrambler_time.svg)

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.5)
        </div>
        """
        pass

    def scramble(self, vt0, vt1):
        r"""
        Embaralha os vetores $v_t^{(0)}$ e $v_t^{(1)}$, retornando os vetores $X_n$ e $Y_n$ embaralhados. O processo de embaralhamento é dado pela expressão abaixo.

        \begin{equation}
            X_n = \begin{cases}
            A, & \text{se } n = 0 \pmod{3} \\
            B, & \text{se } n = 1 \pmod{3} \\
            C, & \text{se } n = 2 \pmod{3}
            \end{cases} \quad
            Y_n = \begin{cases}
            A, & \text{se } n = 0 \pmod{3} \\
            B, & \text{se } n = 1 \pmod{3} \\
            C, & \text{se } n = 2 \pmod{3}
            \end{cases}
        \end{equation}

        Sendo: 
            - $X_n$ e $Y_n$: Vetores de saída embaralhados.
            - $A$, $B$ e $C$: Combinação de bits dos vetores de entrada $v_t^{(0)}$ e $v_t^{(1)}$.
            - $n$: Indice do bit a ser embaralhado.

        O processo de embaralhamento é ilustrado pelo diagrama de blocos abaixo. 

        ![pageplot](../assets/embaralhador.svg)
        
        Args:
            vt0 (np.ndarray): Vetor de entrada $v_t^{(0)}$.
            vt1 (np.ndarray): Vetor de entrada $v_t^{(1)}$.

        Returns:
            X_scrambled (np.ndarray): Vetor $X_n$ embaralhado.
            Y_scrambled (np.ndarray): Vetor $Y_n$ embaralhado.

        Raises:
            AssertionError: Se os vetores X e Y não tiverem o mesmo comprimento.
        """
        assert len(vt0) == len(vt1), "Vetores X e Y devem ter o mesmo comprimento"
        X_scrambled = []
        Y_scrambled = []

        for i in range(0, len(vt0), 3):
            x_blk = vt0[i:i+3]
            y_blk = vt1[i:i+3]
            n = len(x_blk)

            if n == 3:
                # Embaralhamento do bloco [x1, x2, x3], [y1, y2, y3]
                x1, x2, x3 = x_blk
                y1, y2, y3 = y_blk
                X_scrambled += [y1, x2, y2]
                Y_scrambled += [x1, x3, y3]
            elif n == 2:
                # Embaralhamento do bloco [x1, x2], [y1, y2]
                x1, x2 = x_blk
                y1, y2 = y_blk
                X_scrambled += [y1, x2]
                Y_scrambled += [x1, y2]
            elif n == 1:
                # Embaralhamento do bloco [x1], [y1]
                x1 = x_blk[0]
                y1 = y_blk[0]
                X_scrambled += [y1]
                Y_scrambled += [x1]

        return X_scrambled, Y_scrambled

    def descramble(self, X_prime, Y_prime):
        r"""
        Desembaralha os vetores $X'_n$ e $Y'_n$ embaralhados, retornando os vetores $v_t^{(0)'}$ e $v_t^{(1)'}$ restaurados. O processo de desembaralhamento
        é dado pela expressão abaixo.

        \begin{equation}
            v_t^{(0)'} = \begin{cases}
            A, & \text{se } n = 0 \pmod{3} \\
            B, & \text{se } n = 1 \pmod{3} \\
            C, & \text{se } n = 2 \pmod{3}
            \end{cases}, \quad
            v_t^{(1)'} = \begin{cases}
            A, & \text{se } n = 0 \pmod{3} \\
            B, & \text{se } n = 1 \pmod{3} \\
            C, & \text{se } n = 2 \pmod{3}
            \end{cases} \text{ .}
            \label{eq:desembaralhador_Y}
        \end{equation}

        Sendo: 
            - $v_t^{(0)'}$ e $v_t^{(1)'}$: Vetores de saida desembaralhados.
            - $A$, $B$ e $C$: Combinação de bits dos vetores de entrada $X'_n$ e $Y'_n$ embaralhados.
            - $n$: Indice do bit a ser embaralhado.

        O processo de desembaralhamento é ilustrado pelo diagrama de blocos abaixo.

        ![pageplot](../assets/desembaralhador.svg)

        Args:
            X_prime (np.ndarray): Vetor $X'_{n}$ embaralhado.
            Y_prime (np.ndarray): Vetor $Y'_{n}$ embaralhado.

        Returns:
            vt0_prime (np.ndarray): Vetor $v_t^{(0)}$ restaurado.
            vt1_prime (np.ndarray): Vetor $v_t^{(1)}$ restaurado.
        
        Raises:
            AssertionError: Se os vetores X e Y não tiverem o mesmo comprimento.
        """
        assert len(X_prime) == len(Y_prime), "Vetores X e Y devem ter o mesmo comprimento"
        vt0_prime = []
        vt1_prime = []

        for i in range(0, len(X_prime), 3):
            x_blk = X_prime[i:i+3]
            y_blk = Y_prime[i:i+3]
            n = len(x_blk)

            if n == 3:
                # Desembaralhamento do bloco [y1, x2, y2], [x1, x3, y3]
                x1, x2, x3 = y_blk[0], x_blk[1], y_blk[1]
                y1, y2, y3 = x_blk[0], x_blk[2], y_blk[2]
                vt0_prime.extend([x1, x2, x3])
                vt1_prime.extend([y1, y2, y3])
            elif n == 2:
                # Desembaralhamento do bloco [y1, x2], [x1, y2]
                x1, x2 = y_blk[0], x_blk[1]
                y1, y2 = x_blk[0], y_blk[1]
                vt0_prime.extend([x1, x2])
                vt1_prime.extend([y1, y2])
            elif n == 1:
                # Desembaralhamento do bloco [y1], [x1]
                x1 = y_blk[0]
                y1 = x_blk[0]
                vt0_prime.append(x1)
                vt1_prime.append(y1)

        return vt0_prime, vt1_prime



if __name__ == "__main__":
    vt0 = np.random.randint(0, 2, 30)
    vt1 = np.random.randint(0, 2, 30)
    idx_vt0 = [f"X{i+1}" for i in range(len(vt0))]
    idx_vt1 = [f"Y{i+1}" for i in range(len(vt1))]

    # Embaralha o conteúdo dos vetores e os indices
    scrambler = Scrambler()
    Xn, Yn = scrambler.scramble(vt0, vt1)
    idx_Xn, idx_Yn = scrambler.scramble(idx_vt0, idx_vt1)

    print("\nSequência original:")
    print("vt0: ", ''.join(str(b) for b in vt0))
    print("vt1: ", ''.join(str(b) for b in vt1))
    print("idx_vt0:", idx_vt0[:12])
    print("idx_vt1:", idx_vt1[:12])

    print("\nSequência embaralhada:")
    print("Xn  :", ''.join(str(int(b)) for b in Xn))
    print("Yn  :", ''.join(str(int(b)) for b in Yn))
    print("idx_Xn: ", idx_Xn[:12])
    print("idx_Yn: ", idx_Yn[:12])

    # Desembaralha o conteúdo dos vetores e os indices
    vt0_prime, vt1_prime = scrambler.descramble(Xn, Yn)
    idx_vt0_prime, idx_vt1_prime = scrambler.descramble(idx_Xn, idx_Yn)

    print("\nVerificação:")
    print("vt0':", ''.join(str(int(b)) for b in vt0_prime))
    print("vt1':", ''.join(str(int(b)) for b in vt1_prime))
    print("idx_vt0': ", idx_vt0_prime[:12])
    print("idx_vt1': ", idx_vt1_prime[:12])
    print("vt0 = vt0': ", np.array_equal(vt0, vt0_prime))
    print("vt1 = vt1': ", np.array_equal(vt1, vt1_prime))
    print("idx_vt0 = idx_vt0': ", np.array_equal(idx_vt0, idx_vt0_prime))
    print("idx_vt1 = idx_vt1': ", np.array_equal(idx_vt1, idx_vt1_prime))

    fig_scrambler, grid_scrambler = create_figure(3, 2, figsize=(16, 9))

    BitsPlot(
        fig_scrambler, grid_scrambler, (0, 0),
        bits_list=[vt0],
        sections=[("$v_t^{0}$", len(vt0))],
        colors=["darkgreen"],
        ylabel="Original"
    ).plot()

    BitsPlot(
        fig_scrambler, grid_scrambler, (0, 1),
        bits_list=[vt1],
        sections=[("$v_t^{1}$", len(vt1))],
        colors=["navy"]
    ).plot()

    BitsPlot(
        fig_scrambler, grid_scrambler, (1, 0),
        bits_list=[Xn],
        sections=[("$X_n$", len(Xn))],
        colors=["darkgreen"],
        ylabel="Embaralhado"
    ).plot()

    BitsPlot(
        fig_scrambler, grid_scrambler, (1, 1),
        bits_list=[Yn],
        sections=[("$Y_n$", len(Yn))],
        colors=["navy"]
    ).plot()

    BitsPlot(
        fig_scrambler, grid_scrambler, (2, 0),
        bits_list=[vt0_prime],
        sections=[("$v_t^{0}$", len(vt0_prime))],
        colors=["darkgreen"],
        ylabel="Restaurado", xlabel="Index de Bit"
    ).plot()

    BitsPlot(
        fig_scrambler, grid_scrambler, (2, 1),
        bits_list=[vt1_prime],
        sections=[("$v_t^{1}$", len(vt1_prime))],
        colors=["navy"],
        xlabel="Index de Bit"
    ).plot()

    fig_scrambler.tight_layout()
    save_figure(fig_scrambler, "example_scrambler_time.pdf")


# """
# Implementação de um transmissor PTT-A3 com seus componentes.

# Autor: Arthur Cadore
# Data: 16-08-2025
# """

import numpy as np
from .formatter import Formatter
from .convolutional import EncoderConvolutional
from .datagram import Datagram
from .modulator import Modulator
from .preamble import Preamble
from .scrambler import Scrambler
from .multiplexer import Multiplexer
from .encoder import Encoder
from .data import ExportData
from .plotter import create_figure, save_figure, BitsPlot, ImpulseResponsePlot, TimePlot, FrequencyPlot, ConstellationPlot, PhasePlot, SymbolsPlot

class Transmitter:
    def __init__(self, fc=4000, fs=128_000, Rb=400, carrier_length=0.082, preamble="2BEEEEBF", channel_encode=("nrz", "man"), G=np.array([[0b1111001, 0b1011011]]), output_print=True, output_plot=True):
        r"""
        Classe que encapsula todo o processo de transmissão no padrão PTT-A3. A estrutura do transmissor é representada pelo diagrama de blocos abaixo.
    
        Args:
            fc (float): Frequência da portadora em Hz. 
            fs (float): Frequência de amostragem em Hz. 
            Rb (float): Taxa de bits em bps.
            carrier_length (float): Comprimento do prefixo em segundos.
            preamble (str): String de preâmbulo em hex.
            channel_encode (tuple): Tupla com o tipo de codificação dos canais I e Q respectivamente.
            G (np.ndarray): Matriz de geração para codificação convolucional.
            output_print (bool): Se `True`, imprime os vetores intermediários no console.
            output_plot (bool): Se `True`, gera e salva os gráficos dos processos intermediários.

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1 e 3.2)
        </div>
        """

        # Validar os valores de channel_encode
        valid_encodings = ["nrz", "man"]
        if channel_encode[0] not in valid_encodings or channel_encode[1] not in valid_encodings:
            raise ValueError("Os tipos de codificação devem ser 'nrz' ou 'manchester'.")

        # Parâmetros
        self.fc = fc
        self.fs = fs
        self.Rb = Rb
        self.output_print = output_print
        self.output_plot = output_plot
        self.prefix_duration = carrier_length

        # Parâmetros fixos
        self.alpha = 0.8
        self.span = 12
        self.cI_encoder = "nrz"
        self.cQ_encoder = "nrz"

        # Codificação I e Q
        self.cI_type = channel_encode[0]
        self.cQ_type = channel_encode[1]

        # Mapeamento das configurações de codificação
        encoding_params = {
            "nrz": {"format": "RRC", "bits_per_symbol": 1, "Rb_multiplier": 1},
            "man": {"format": "Manchester", "bits_per_symbol": 2, "Rb_multiplier": 2}
        }

        # Parâmetros para o canal I e Q
        cI_params = encoding_params[self.cI_type]
        self.cI_format = cI_params["format"]
        self.cI_bits_per_symbol = cI_params["bits_per_symbol"]
        self.cI_Rb = self.Rb
        cQ_params = encoding_params[self.cQ_type]
        self.cQ_format = cQ_params["format"]
        self.cQ_bits_per_symbol = cQ_params["bits_per_symbol"]
        self.cQ_Rb = self.Rb

        # Submodulos
        self.encoder = EncoderConvolutional(G=G)
        self.scrambler = Scrambler()
        self.preamble = Preamble(preamble_hex=preamble)
        self.multiplexer = Multiplexer()
        self.c_encoderI = Encoder(self.cI_encoder)
        self.c_encoderQ = Encoder(self.cQ_encoder)
        self.formatterI = Formatter(fs=self.fs, Rb=self.cI_Rb, type=self.cI_format, channel="I", bits_per_symbol=self.cI_bits_per_symbol, prefix_duration=self.prefix_duration, alpha=self.alpha, span=self.span)
        self.formatterQ = Formatter(fs=self.fs, Rb=self.cQ_Rb, type=self.cQ_format, channel="Q", bits_per_symbol=self.cQ_bits_per_symbol, prefix_duration=self.prefix_duration, alpha=self.alpha, span=self.span)
        self.modulator = Modulator(fc=self.fc, fs=self.fs)

    def prepare_datagram(self, datagram: Datagram):
        r"""
        Gera o datagrama para transmissão, retornando o vetor de bits $u_t$.

        Returns:
            ut (np.ndarray): Vetor de bits do datagrama.

        Example:
            ![pageplot](assets/transmitter_datagram_time.svg)
        """

        ut = datagram.streambits

        if self.output_print:
            print("\n ==== MONTAGEM DATAGRAMA ==== \n")
            print(datagram.parse_datagram())
            print("\nut:", ''.join(map(str, ut)))

        if self.output_plot:
            fig_datagram, grid = create_figure(1, 1, figsize=(16, 5))

            BitsPlot(
                fig_datagram, grid, (0, 0),
                bits_list=[datagram.msglength, 
                           datagram.pcdid, 
                           datagram.blocks, 
                           datagram.tail],
                sections=[("Message Length", len(datagram.msglength)),
                          ("PCD ID", len(datagram.pcdid)),
                          ("Dados de App.", len(datagram.blocks)),
                          ("Tail", len(datagram.tail))],
                colors=["green", "orange", "red", "blue"],
                xlabel="Index de Bit"
            ).plot()

            fig_datagram.tight_layout()
            save_figure(fig_datagram, "transmitter_datagram_time.pdf")

        return ut

    def encode_convolutional(self, ut):
        r"""
        Codifica o vetor de bits $u_t$ usando codificação convolucional, retornando os vetores de bits $v_t^{(0)}$ e $v_t^{(1)}$.

        Args:
            ut (np.ndarray): Vetor de bits a ser codificado.

        Returns:
            vt0 (np.ndarray): Saída do canal I.
            vt1 (np.ndarray): Saída do canal Q.

        Example:
            ![pageplot](assets/transmitter_conv_time.svg)
        """

        vt0, vt1 = self.encoder.encode(ut)

        if self.output_print:
            print("\n ==== CODIFICADOR CONVOLUCIONAL ==== \n")
            print("vt0:", ''.join(map(str, vt0)))
            print("vt1:", ''.join(map(str, vt1)))

        if self.output_plot:
            fig_conv, grid_conv = create_figure(3, 1, figsize=(16, 9))

            BitsPlot(
                fig_conv, grid_conv, (0, 0),
                bits_list=[ut],
                sections=[("$u_t$", len(ut))],
                colors=["darkred"],
                ylabel="$u_t$",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_conv, grid_conv, (1, 0),
                bits_list=[vt0],
                sections=[("$v_t^{(0)}$", len(vt0))],
                colors=["darkgreen"],
                ylabel="$v_t^{(0)}$",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_conv, grid_conv, (2, 0),
                bits_list=[vt1],
                sections=[("$v_t^{(1)}$", len(vt1))],
                colors=["navy"],
                ylabel="$v_t^{(1)}$", 
                xlim=(0, 60),
                xlabel="Index de Bit"
            ).plot()

            fig_conv.tight_layout()
            save_figure(fig_conv, "transmitter_conv_time.pdf")       
        return vt0, vt1

    def scramble(self, vt0, vt1):
        r"""
        Embaralha os vetores de bits $v_t^{(0)}$ e $v_t^{(1)}$, criando os vetores $X_n$ e $Y_n$ embaralhados.

        Args:
            vt0 (np.ndarray): Vetor de bits do canal I.
            vt1 (np.ndarray): Vetor de bits do canal Q.

        Returns:
            Xn (np.ndarray): Vetor embaralhado do canal I.
            Yn (np.ndarray): Vetor embaralhado do canal Q.

        Example:
            ![pageplot](assets/transmitter_scrambler_time.svg)
        """

        X, Y = self.scrambler.scramble(vt0, vt1)

        if self.output_print:
            print("\n ==== EMBARALHADOR ==== \n")
            print("Xn:", ''.join(map(str, X)))
            print("Yn:", ''.join(map(str, Y)))
            
        if self.output_plot:
            fig_scrambler, grid_scrambler = create_figure(4, 1, figsize=(16, 9))

            BitsPlot(
                fig_scrambler, grid_scrambler, (0, 0),
                bits_list=[vt0],
                sections=[("$v_t^{0}$", len(vt0))],
                colors=["darkgreen"],
                xlim=(0, 60),
                ylabel="Original"
            ).plot()

            BitsPlot(
                fig_scrambler, grid_scrambler, (1, 0),
                bits_list=[X],
                sections=[("$X_n$", len(X))],
                colors=["darkgreen"],
                ylabel="Embaralhado", 
                xlim=(0, 60),
            ).plot()

            BitsPlot(
                fig_scrambler, grid_scrambler, (2, 0),
                bits_list=[vt1],
                sections=[("$v_t^{1}$", len(vt1))],
                colors=["navy"],
                xlim=(0, 60),
                ylabel="Original"
            ).plot()

            BitsPlot(
                fig_scrambler, grid_scrambler, (3, 0),
                bits_list=[Y],
                sections=[("$Y_n$", len(Y))],
                colors=["navy"], 
                ylabel="Embaralhado", 
                xlabel="Index de Bit",
                xlim=(0, 60),
            ).plot()

            fig_scrambler.tight_layout()
            save_figure(fig_scrambler, "transmitter_scrambler_time.pdf")

        return X, Y

    def generate_preamble(self):
        r"""
        Gera os vetores de preâmbulo $S_I$ e $S_Q$.

        Returns:
            sI (np.ndarray): Vetor do preâmbulo do canal I.
            sQ (np.ndarray): Vetor do preâmbulo do canal Q.

        Example:
            ![pageplot](assets/transmitter_preamble_time.svg)
        """
        
        sI, sQ = self.preamble.generate_preamble()

        if self.output_print:
            print("\n ==== MONTAGEM PREAMBULO ==== \n")
            print("sI:", ''.join(map(str, sI)))
            print("sQ:", ''.join(map(str, sQ)))

        if self.output_plot:
            fig_preamble, grid_preamble = create_figure(2, 1, figsize=(16, 6))

            BitsPlot(
                fig_preamble, grid_preamble, (0,0),
                bits_list=[sI],
                sections=[("$S_I$", len(sI))],
                colors=["darkgreen"],
                ylabel="Canal $I$"
            ).plot()
            
            BitsPlot(
                fig_preamble, grid_preamble, (1,0),
                bits_list=[sQ],
                sections=[("$S_Q$", len(sQ))],
                colors=["navy"], 
                xlabel="Index de Bit", 
                ylabel="Canal $Q$"
            ).plot()

            fig_preamble.tight_layout()
            save_figure(fig_preamble, "transmitter_preamble_time.pdf")

        return sI, sQ

    def multiplex(self, sI, sQ, X, Y):
        r"""
        Multiplexa os vetores de preâmbulo $S_I$ e $S_Q$ com os vetores de dados $X$ e $Y$, retornando os vetores multiplexados $X_n$ e $Y_n$.

        Args:
            sI (np.ndarray): Vetor do preâmbulo do canal I.
            sQ (np.ndarray): Vetor do preâmbulo do canal Q.
            X (np.ndarray): Vetor de dados do canal I.
            Y (np.ndarray): Vetor de dados do canal Q.
        
        Returns:
            Xn (np.ndarray): Vetor multiplexado do canal I.
            Yn (np.ndarray): Vetor multiplexado do canal Q.

        Example:
            ![pageplot](assets/transmitter_mux_time.svg)
        """

        Xn, Yn = self.multiplexer.concatenate(sI, sQ, X, Y)

        if self.output_print:
            print("\n ==== MULTIPLEXADOR ==== \n")
            print("Xn:", ''.join(map(str, Xn)))
            print("Yn:", ''.join(map(str, Yn)))

        if self.output_plot:
            fig_mux, grid_mux = create_figure(2, 1, figsize=(16, 9))

            BitsPlot(
                fig_mux, grid_mux, (0,0),
                bits_list=[sI, X],
                sections=[("Preambulo $S_I$", len(sI)),
                          ("Canal I $(X_n)$", len(X))],
                colors=["darkred", "darkgreen"],
                ylabel="Canal $I$",
                xlim=(0, 60),
            ).plot()

            BitsPlot(
                fig_mux, grid_mux, (1,0),
                bits_list=[sQ, Y],
                sections=[("Preambulo $S_Q$", len(sQ)),
                          ("Canal Q $(Y_n)$", len(Y))],
                colors=["darkred", "navy"],
                xlabel="Index de Bit", 
                ylabel="Canal $Q$",
                xlim=(0, 60),
            ).plot()

            fig_mux.tight_layout()
            save_figure(fig_mux, "transmitter_mux_time.pdf")   
        return Xn, Yn

    def encode_channels(self, Xn, Yn):
        r"""
        Codifica os vetores dos canais $X_n$ e $Y_n$ usando $NRZ$ e $Manchester$, respectivamente, retornando os vetores de sinal codificados $X_{NRZ}$ e $Y_{MAN}$.

        Args:
            Xn (np.ndarray): Vetor do canal $X_n$ a ser codificado.
            Yn (np.ndarray): Vetor do canal $Y_n$ a ser codificado.
        
        Returns:
            Xnrz (np.ndarray): Vetor de sinal codificado do canal I $NRZ$. 
            Yman (np.ndarray): Vetor de sinal codificado do canal Q $Manchester$. 

        Example:
            ![pageplot](assets/transmitter_encoder_time.svg)
        """

        Xi = self.c_encoderI.encode(Xn)
        Yq = self.c_encoderQ.encode(Yn)

        if self.output_print:
            print("\n ==== CODIFICAÇÃO DE LINHA ==== \n")
            print("Xi:", ' '.join(f"{x:+d}" for x in Xi[:40]),"...")
            print("Yq:", ' '.join(f"{y:+d}" for y in Yq[:40]),"...")

        if self.output_plot:
            fig_encoder, grid = create_figure(4, 1, figsize=(16, 9))

            BitsPlot(
                fig_encoder, grid, (0, 0),
                bits_list=[Xn],
                sections=[("$X_n$", len(Xn))],
                colors=["darkgreen"],
                xlabel="Index de Bit", 
                ylabel="$X_n$", 
                xlim=(0, 60)
            ).plot()

            SymbolsPlot(
                fig_encoder, grid, (1, 0),
                symbols_list=[Xi],
                samples_per_symbol=1,
                colors=["darkgreen"],
                xlabel="Index de Simbolo",
                xlim=(0, 60),
                ylabel="$X_{NRZ}[n]$"
            ).plot()

            BitsPlot(
                fig_encoder, grid, (2, 0),
                bits_list=[Yn],
                sections=[("$Y_n$", len(Yn))],
                colors=["navy"],
                xlabel="Index de Bit", 
                ylabel="$Y_n$", 
                xlim=(0, 60),
            ).plot()

            SymbolsPlot(
                fig_encoder, grid, (3, 0),
                symbols_list=[Yq],
                samples_per_symbol=1,
                colors=["navy"],
                xlabel="Index de Simbolo",
                xlim=(0, 60),
                ylabel="$Y_{MAN}[n]$"
            ).plot()

            fig_encoder.tight_layout()
            save_figure(fig_encoder, "transmitter_encoder_time.pdf")

        return Xi, Yq

    def format_signals(self, Xi, Yq):
        r"""
        Formata os vetores de sinal codificados $X_{NRZ}$ e $Y_{MAN}$ usando filtro RRC, retornando os vetores formatados $d_I$ e $d_Q$.

        Args:
            Xnrz (np.ndarray): Vetor do canal $X_{NRZ}$ a ser formatado.
            Yman (np.ndarray): Vetor do canal $Y_{MAN}$ a ser formatado.
        
        Returns:
            dI (np.ndarray): Vetor formatado do canal I, $d_I$.
            dQ (np.ndarray): Vetor formatado do canal Q, $d_Q$.

        Example:
            - Tempo: ![pageplot](assets/transmitter_formatter_time.svg)
            - Frequência: ![pageplot](assets/transmitter_formatter_freq.svg)
        """

        dI = self.formatterI.apply_format(Xi)
        dQ = self.formatterQ.apply_format(Yq)
        
        if self.output_print:
            print("\n ==== FORMATADOR ==== \n")
            print("dI:", ''.join(map(str, dI[:5])),"...")
            print("dQ:", ''.join(map(str, dQ[:5])),"...")
            print("Prefix Duration:", self.prefix_duration)
            
        if self.output_plot:
            fig_format, grid_format = create_figure(2, 2, figsize=(16, 9))

            ImpulseResponsePlot(
                fig_format, grid_format, (0, 0),
                self.formatterI.t_rc, self.formatterI.g,
                t_unit="ms",
                colors="darkorange",
                label="$g(t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            ImpulseResponsePlot(
                fig_format, grid_format, (0, 1),
                self.formatterQ.t_rc, self.formatterQ.g,
                t_unit="ms",
                colors="darkorange",
                label="$g(t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            TimePlot(
                fig_format, grid_format, (1,0),
                t= np.arange(len(dI)) / self.formatterI.fs,
                signals=[dI],
                labels=["$d_I(t)$"],
                title="Canal $I$",
                xlim=(40, 200),
                amp_norm=True,
                colors="darkgreen",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            TimePlot(
                fig_format, grid_format, (1,1),
                t= np.arange(len(dQ)) / self.formatterQ.fs,
                signals=[dQ],
                labels=["$d_Q(t)$"],
                title="Canal $Q$",
                xlim=(40, 200),
                amp_norm=True,
                colors="darkblue",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            fig_format.tight_layout()
            save_figure(fig_format, "transmitter_formatter_time.pdf")

            fig_format_freq, grid_format_freq = create_figure(2, 2, figsize=(16, 9))

            ImpulseResponsePlot(
                fig_format_freq, grid_format_freq, (0, 0),
                self.formatterI.t_rc, self.formatterI.g,
                t_unit="ms",
                colors="darkorange",
                label="$g(t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15), 
                amp_norm=True
            ).plot()

            ImpulseResponsePlot(
                fig_format_freq, grid_format_freq, (0, 1),
                self.formatterQ.t_rc, self.formatterQ.g,
                t_unit="ms",
                colors="darkorange",
                label="$g(t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15), 
                amp_norm=True
            ).plot()

            FrequencyPlot(
                fig_format_freq, grid_format_freq, (1, 0),
                fs=self.fs,
                signal=dI,
                fc=self.fc,
                labels=["$D_I(f)$"],
                title="Canal $I$",
                xlim=(-1.5, 1.5),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_format_freq, grid_format_freq, (1, 1),
                fs=self.fs,
                signal=dQ,
                fc=self.fc,
                labels=["$D_Q(f)$"],
                title="Canal $Q$",
                xlim=(-1.5, 1.5),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_format_freq.tight_layout()
            save_figure(fig_format_freq, "transmitter_formatter_freq.pdf")

        return dI, dQ

    def modulate(self, dI, dQ):
        r"""
        Modula os vetores de sinal $d_I(t)$ e $d_Q(t)$ usando modulação QPSK, retornando o sinal modulado $s(t)$.

        Args:
            dI (np.ndarray): Vetor formatado do canal I, $d_I(t)$.
            dQ (np.ndarray): Vetor formatado do canal Q, $d_Q(t)$.
        
        Returns:
            t (np.ndarray): Vetor de tempo, $t$.
            s (np.ndarray): Sinal modulado, $s(t)$.

        Example:
            - Tempo: ![pageplot](assets/transmitter_modulator_time.svg)
            - Frequência: ![pageplot](assets/transmitter_modulator_freq.svg)
            - Portadora: ![pageplot](assets/transmitter_modulator_portadora.svg)
            - Fase e Constelação: ![pageplot](assets/transmitter_modulator_constellation.svg)
        """

        t, s = self.modulator.modulate(dI, dQ)

        if self.output_print:
            print("\n ==== MODULADOR ==== \n")
            print("s(t):", ''.join(map(str, s[:5])),"...")
            print("t:   ", ''.join(map(str, t[:5])),"...")

        if self.output_plot:
            fig_time, grid = create_figure(2, 1, figsize=(16, 9))

            TimePlot(
                fig_time, grid, (0, 0),
                t=t,
                signals=[dI, dQ],
                labels=["$d_I(t)$", "$d_Q(t)$"],
                title="Componentes $IQ$ - Demoduladas",
                xlim=(40, 200),
                amp_norm=True,
                colors=["darkgreen", "navy"],
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            TimePlot(
                fig_time, grid, (1, 0),
                t=t,
                signals=[s],
                labels=["$s(t)$"],
                title="Sinal Modulado $IQ$",
                xlim=(40, 200),
                amp_norm=True,
                colors="darkred",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            fig_time.tight_layout()
            save_figure(fig_time, "transmitter_modulator_time.pdf")

            fig_freq, grid = create_figure(2, 2, figsize=(16, 9))
            FrequencyPlot(
                fig_freq, grid, (0, 0),
                fs=self.fs,
                signal=dI,
                fc=self.fc,
                labels=["$D_I(f)$"],
                title="Componente I",
                xlim=(-1.5, 1.5),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()
        
            FrequencyPlot(
                fig_freq, grid, (0, 1),
                fs=self.fs,
                signal=dQ,
                fc=self.fc,
                labels=["$D_Q(f)$"],
                title="Componente Q",
                xlim=(-1.5, 1.5),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()
        
            FrequencyPlot(
                fig_freq, grid, (1, slice(0, 2)),
                fs=self.fs,
                signal=s,
                fc=self.fc,
                labels=["$S(f)$"],
                title="Sinal Modulado $IQ$",
                xlim=(0, 8),
                colors="darkred",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()
        
            fig_freq.tight_layout()
            save_figure(fig_freq, "transmitter_modulator_freq.pdf")

            fig_const, grid = create_figure(1, 2, figsize=(16, 8))
            PhasePlot(
                fig_const, grid, (0, 0),
                t=t,
                signals=[dI, dQ],
                labels=["Fase $I + jQ$"],
                title="Fase $I + jQ$",
                xlim=(40, 320),
                colors=["darkred"],
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            ConstellationPlot(
                fig_const, grid, (0, 1),
                dI=dI[:40000:5],
                dQ=dQ[:40000:5],
                xlim=(-1.4, 1.4),
                ylim=(-1.4, 1.4),
                rms_norm=True,
                show_ideal_points=False,
                title="Constelação $IQ$",
                colors=["darkred"],
                style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_const.tight_layout()
            save_figure(fig_const, "transmitter_modulator_constellation.pdf") 

            fig_portadora, grid = create_figure(1, 2, figsize=(16, 8))
            FrequencyPlot(
                fig_portadora, grid, (0, 0),
                fs=self.fs,
                signal=s[0:(int(round(0.082 * self.fs)))],
                fc=self.fc,
                labels=["$S(f)$"],
                title="Portadora Pura - $0$ a $80$ms",
                xlim=(-10, 10),
                colors="darkred",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_portadora, grid, (0, 1),
                fs=self.fs,
                signal=s[(int(round(0.082 * self.fs))):],
                fc=self.fc,
                labels=["$S(f)$"],
                title="Sinal Modulado - $80$ms em diante",
                xlim=(-10, 10),
                colors="darkred",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_portadora.tight_layout()
            save_figure(fig_portadora, "transmitter_modulator_portadora.pdf")

        return t, s

    def transmit(self, datagram: Datagram):
        r"""
        Executa toda a cadeia de transmissão para um datagrama, retornando o sinal modulado $s(t)$ e o vetor de tempo $t$.

        Args:
            datagram (Datagram): Instância do datagrama a ser transmitido.

        Returns:
            t (np.ndarray): Vetor de tempo, $t$.
            s (np.ndarray): Sinal modulado, $s(t)$.
        """
        ut = self.prepare_datagram(datagram)
        vt0, vt1 = self.encode_convolutional(ut)
        X, Y = self.scramble(vt0, vt1)
        sI, sQ = self.generate_preamble()
        Xn, Yn = self.multiplex(sI, sQ, X, Y)
        Xnrz, Yman = self.encode_channels(Xn, Yn)
        dI, dQ = self.format_signals(Xnrz, Yman)
        t, s = self.modulate(dI, dQ)
        return t, s


if __name__ == "__main__":

    # Cria uma instância de transmissor
    transmitter = Transmitter(output_print=True, output_plot=True)

    datagram1 = Datagram(pcdnum=1234, numblocks=1, seed=10)
    datagram2 = Datagram(pcdnum=1234, numblocks=1, seed=10)

    # Transmite o datagrama 1
    t1, s1 = transmitter.transmit(datagram1)
    t2, s2 = transmitter.transmit(datagram2)

    # Verifica se os vetores são iguais
    if np.array_equal(s1, s2):
        print("S1 == S2")
    else:
        print("S1 != S2")

    # Exporta os dados
    ExportData([s1, t1], "transmitter_st").save()




# """
# Implementação de um receptor PTT-A3 com seus componentes.

# Autor: Arthur Cadore
# Data: 16-08-2025
# """

import numpy as np
from .datagram import Datagram
from .modulator import Modulator
from .scrambler import Scrambler
from .encoder import Encoder
from .transmitter import Transmitter
from .noise import NoiseEBN0
from .lowpassfilter import LPF
from .matchedfilter import MatchedFilter
from .sampler import Sampler
from .convolutional import DecoderViterbi
from .synchronizer import Synchronizer
from .channel import Channel
from .plotter import save_figure, create_figure, TimePlot, FrequencyPlot, ImpulseResponsePlot, SampledSignalPlot, BitsPlot, PhasePlot, ConstellationPlot, FrequencyResponsePlot, SincronizationPlot, CorrelationPlot, SymbolsPlot

class Receiver:
    def __init__(self, fs=128_000, Rb=400, fc=None, lpf_cutoff=600, preamble="2BEEEEBF", channel_encode=("nrz", "man"), G=np.array([[0b1111001, 0b1011011]]), output_print=True, output_plot=True):
        r"""
        Classe que encapsula todo o processo de recepção no padrão ARGOS-3. A estrutura do receptor é representada pelo diagrama de blocos abaixo.

        Args:
            fs (int): Frequência de amostragem em Hz.
            Rb (int): Taxa de bits em bps.
            fc (int): Frequência de portadora em Hz.
            lpf_cutoff (int): Frequência de corte do filtro passa-baixa em Hz.
            preamble (str): String de preâmbulo em hex.
            channel_encode (tuple): Tupla com o tipo de codificação dos canais I e Q respectivamente.
            G (np.ndarray): Matriz de geração para codificação convolucional.
            output_print (bool): Se `True`, imprime os vetores intermediários no console. 
            output_plot (bool): Se `True`, gera e salva os gráficos dos processos intermediários.

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-2097-CNES (seção 3.1 e 3.2)
        </div>
        """

        # Validar os valores de channel_encode
        valid_encodings = ["nrz", "man"]
        if channel_encode[0] not in valid_encodings or channel_encode[1] not in valid_encodings:
            raise ValueError("Os tipos de codificação devem ser 'nrz' ou 'manchester'.")

        # Parâmetros
        self.fs = fs
        self.Rb = Rb
        self.fc = fc
        self.lpf_cutoff = lpf_cutoff
        self.output_print = output_print
        self.output_plot = output_plot
        self.preamble = preamble
        self.G = G

        # Parâmetros fixos
        self.alpha = 0.8
        self.span = 24
        self.lpf_order = 6
        self.delayI = 0
        self.delayQ = 0
        self.cI_encoder = "nrz"
        self.cQ_encoder = "nrz"

        # Codificação I e Q
        self.cI_type = channel_encode[0]
        self.cQ_type = channel_encode[1]

        # Mapeamento das configurações de codificação
        encoding_params = {
            "nrz": {"format": "RRC", "bits_per_symbol": 1, "Rb_multiplier": 1, "matched": "RRC-Inverted"},
            "man": {"format": "Manchester", "bits_per_symbol": 2, "Rb_multiplier": 2, "matched": "Manchester-Inverted"}
        }

        # Parâmetros para o canal I e Q
        cI_params = encoding_params[self.cI_type]
        self.cI_format = cI_params["format"]
        self.cI_bits_per_symbol = cI_params["bits_per_symbol"]
        self.cI_Rb = self.Rb
        self.cI_matched = cI_params["matched"]
        cQ_params = encoding_params[self.cQ_type]
        self.cQ_format = cQ_params["format"]
        self.cQ_bits_per_symbol = cQ_params["bits_per_symbol"]
        self.cQ_Rb = self.Rb
        self.cQ_matched = cQ_params["matched"]


        # Submodulos
        self.demodulator = Modulator(fc=self.fc, fs=self.fs)
        self.lpf = LPF(cut_off=self.lpf_cutoff, order=self.lpf_order, fs=self.fs, type="butter")
        self.matched_filterI = MatchedFilter(alpha=self.alpha, fs=self.fs, Rb=self.Rb, span=self.span, type="RRC-Inverted", channel="I", bits_per_symbol=1)
        self.matched_filterQ = MatchedFilter(alpha=self.alpha, fs=self.fs, Rb=self.Rb, span=self.span, type="Manchester-Inverted", channel="Q", bits_per_symbol=2)
        self.synchronizerI = Synchronizer(fs=self.fs, Rb=self.Rb, sync_word=self.preamble, sync_window=(0, 0.2))
        self.synchronizerQ = Synchronizer(fs=self.fs, Rb=self.Rb, sync_word=self.preamble, sync_window=(0, 0.2))
        self.samplerI = Sampler(fs=self.fs, Rb=self.Rb, delay=self.delayI)
        self.samplerQ = Sampler(fs=self.fs, Rb=self.Rb, delay=self.delayQ)
        self.decoderI = Encoder("nrz")
        self.decoderQ = Encoder("nrz")
        self.unscrambler = Scrambler()
        self.conv_viterbi = DecoderViterbi(G=self.G)

    def demodulate(self, s, t):
        r"""normalmente
        Demodula o sinal $s'(t)$ com ruído recebido, recuperando os sinais $x'_{I}(t)$ e $y'_{Q}(t)$.

        Args:
            s (np.ndarray): Sinal $s'(t)$ a ser demodulado.
            t (np.ndarray): Vetor de tempo.

        Returns:
            xI_prime (np.ndarray): Sinal $x'_{I}(t)$ demodulado.
            yQ_prime (np.ndarray): Sinal $y'_{Q}(t)$ demodulado.
        
        Example:
            - Tempo: ![pageplot](assets/receiver_demodulator_time.svg)
            - Frequência: ![pageplot](assets/receiver_demodulator_freq.svg)
        """

        xI_prime, yQ_prime, phase_estimate, original_phase = self.demodulator.demodulate(s)

        if self.output_print:
            print("\n ==== DEMODULADOR ==== \n")
            print("x'I(t):", ''.join(map(str, xI_prime[:5])),"...")
            print("y'Q(t):", ''.join(map(str, yQ_prime[:5])),"...")
            print("\n")
            print("Erro médio estimado:", np.mean(np.angle(np.exp(1j*(phase_estimate - original_phase)))))

        if self.output_plot:
            fig_time, grid = create_figure(2, 1, figsize=(16, 9))

            TimePlot(
                fig_time, grid, (0, 0),
                t=t,
                signals=[s],
                labels=[r"$s(t)$ + AWGN"],
                title=r"Sinal Modulado + Ruído $Eb/N_0$ 20 $dB$",
                xlim=(40, 200),
                amp_norm=True,
                colors="darkred",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            TimePlot(
                fig_time, grid, (1, 0),
                t=t,
                signals=[xI_prime, yQ_prime],
                labels=[r"$xI'(t)$", r"$yQ'(t)$"],
                title=r"Componentes $IQ$ - Demoduladas",
                xlim=(40, 200),
                amp_norm=True,
                colors=["darkgreen", "navy"],
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            fig_time.tight_layout()
            save_figure(fig_time, "receiver_demodulator_time.pdf")

            fig_freq, grid = create_figure(3, 1, figsize=(16, 9))

            FrequencyPlot(
                fig_freq, grid, (0, 0),
                fs=self.fs,
                signal=s,
                fc=self.fc,
                labels=[r"$S(f)$"],
                title=r"Sinal Modulado $IQ$",
                xlim=(-10, 10),
                colors="darkred",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_freq, grid, (1, 0),
                fs=self.fs,
                signal=xI_prime,
                fc=self.fc,
                labels=[r"$X_I'(f)$"],
                title=r"Componente $I$ - Demodulada",
                xlim=(-10, 10),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_freq, grid, (2, 0),
                fs=self.fs,
                signal=yQ_prime,
                fc=self.fc,
                labels=[r"$Y_Q'(f)$"],
                title=r"Componente $Q$ - Demodulada",
                xlim=(-10, 10),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_freq.tight_layout()
            save_figure(fig_freq, "receiver_demodulator_freq.pdf")

        return xI_prime, yQ_prime
    
    def lowpassfilter(self, xI_prime, yQ_prime, t):
        r"""
        Aplica o filtro passa-baixa com resposta ao impuslo $h(t)$ aos sinais $x'_{I}(t)$ e $y'_{Q}(t)$, retornando os sinais filtrados $d'_{I}(t)$ e $d'_{Q}(t)$.

        Args:
            xI_prime (np.ndarray): Sinal $x'_{I}(t)$ a ser filtrado.
            yQ_prime (np.ndarray): Sinal $y'_{Q}(t)$ a ser filtrado.
            t (np.ndarray): Vetor de tempo.

        Returns:
            dI_prime (np.ndarray): Sinal $d'_{I}(t)$ filtrado.
            dQ_prime (np.ndarray): Sinal $d'_{Q}(t)$ filtrado.

        Example:
            - Tempo: ![pageplot](assets/receiver_lpf_time.svg)
            - Frequência: ![pageplot](assets/receiver_lpf_freq.svg)
        """

        impulse_response, t_impulse = self.lpf.calc_impulse_response()

        dI_prime = self.lpf.apply_filter(xI_prime)
        dQ_prime = self.lpf.apply_filter(yQ_prime)

        if self.output_print:
            print("\n ==== FILTRAGEM PASSA-BAIXA ==== \n")
            print("d'I(t):", ''.join(map(str, dI_prime[:5])),"...")
            print("d'Q(t):", ''.join(map(str, dQ_prime[:5])),"...")
        
        if self.output_plot:
            fig_signal, grid_signal = create_figure(2, 2, figsize=(16, 9))

            ImpulseResponsePlot(
                fig_signal, grid_signal, (0, slice(0, 2)),
                t_impulse, impulse_response,
                t_unit="ms",
                colors="darkorange",
                label=r"$h(t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(0, 8),
                amp_norm=True,
            ).plot()

            TimePlot(
                fig_signal, grid_signal, (1, 0),
                t=t, 
                signals=[dI_prime],
                labels=["$d_I'(t)$"],  
                title="Sinal filtrado - Componente $I$",
                xlim=(40, 200),
                ylim=(-1, 1),
                amp_norm=True,
                colors="darkgreen"
            ).plot()

            TimePlot(
                fig_signal, grid_signal, (1, 1),
                t=t, 
                signals=[dQ_prime],
                labels=["$d_Q'(t)$"],
                title="Sinal filtrado - Componente $Q$",
                xlim=(40, 200),
                ylim=(-1, 1),
                amp_norm=True,
                colors="navy"
            ).plot()

            fig_signal.tight_layout()
            save_figure(fig_signal, "receiver_lpf_time.pdf")

            fig_freq, grid_freq = create_figure(3, 2, figsize=(16, 9))

            FrequencyResponsePlot(
                fig_freq, grid_freq, (0, slice(0, 2)),
                self.lpf.b,
                self.lpf.a,
                self.fs,
                f_cut=self.lpf_cutoff,
                xlim=(0, 3*self.lpf_cutoff),
            ).plot()

            FrequencyPlot(
                fig_freq, grid_freq, (1, 0), 
                fs=self.fs,
                signal=xI_prime,
                fc=self.fc,
                labels=[r"$X_I'(f)$"],
                title=r"Componente $I$ - Demodulada",
                xlim=(-10, 10),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_freq, grid_freq, (1, 1), 
                fs=self.fs,
                signal=yQ_prime,
                fc=self.fc,
                labels=[r"$Y_Q'(f)$"],
                title=r"Componente $Q$ - Demodulada",
                xlim=(-10, 10),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_freq, grid_freq, (2, 0), 
                fs=self.fs,
                signal=dI_prime,
                fc=self.fc,
                labels=[r"$d_I'(f)$"],
                title=r"Componente $I$ - Filtrada",
                xlim=(-10, 10),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_freq, grid_freq, (2, 1), 
                fs=self.fs,
                signal=dQ_prime,
                fc=self.fc,
                labels=[r"$d_Q'(f)$"],
                title=r"Componente $Q$ - Filtrada",
                xlim=(-10, 10),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_freq.tight_layout()
            save_figure(fig_freq, "receiver_lpf_freq.pdf")


        return dI_prime, dQ_prime

    def matchedfilter(self, dI_prime, dQ_prime, t):
        r"""
        Aplica o filtro casado com resposta ao impuslo $g(-t)$ aos sinais $d'_{I}(t)$ e $d'_{Q}(t)$, retornando os sinais filtrados $I'(t)$ e $Q'(t)$.

        Args:
            dI_prime (np.ndarray): Sinal $d'_{I}(t)$ a ser filtrado.
            dQ_prime (np.ndarray): Sinal $d'_{Q}(t)$ a ser filtrado.
            t (np.ndarray): Vetor de tempo.

        Returns:
            It_prime (np.ndarray): Sinal $I'(t)$ filtrado.
            Qt_prime (np.ndarray): Sinal $Q'(t)$ filtrado.

        Example:
            - Tempo: ![pageplot](assets/receiver_mf_time.svg)
            - Frequência: ![pageplot](assets/receiver_mf_freq.svg)
        """

        It_prime = self.matched_filterI.apply_filter(dI_prime)
        Qt_prime = self.matched_filterQ.apply_filter(dQ_prime)

        if self.output_print:
            print("\n ==== FILTRAGEM CASADA ==== \n")
            print("I'(t):", ''.join(map(str, It_prime[:5])),"...")
            print("Q'(t):", ''.join(map(str, Qt_prime[:5])),"...")

        if self.output_plot:
            fig_matched, grid_matched = create_figure(2, 2, figsize=(16, 9))

            ImpulseResponsePlot(
                fig_matched, grid_matched, (0, 0),
                self.matched_filterI.t_rc, self.matched_filterI.g_inverted,
                t_unit="ms",
                colors="darkorange",
                label=r"$g(-t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            ImpulseResponsePlot(
                fig_matched, grid_matched, (0, 1),
                self.matched_filterQ.t_rc, self.matched_filterQ.g_inverted,
                t_unit="ms",
                colors="darkorange",
                label=r"$g(-t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            TimePlot(
                fig_matched, grid_matched, (1, 0),
                t,
                It_prime,
                labels=[r"$I'(t)$"],
                title=r"Sinal filtrado - Componente $I$",
                xlim=(40, 200),
                amp_norm=True,
                colors="darkgreen"
            ).plot()

            TimePlot(
                fig_matched, grid_matched, (1, 1),
                t,
                Qt_prime,
                labels=[r"$Q'(t)$"],
                title=r"Sinal filtrado - Componente $Q$",
                xlim=(40, 200),
                amp_norm=True,
                colors="navy"
            ).plot()

            fig_matched.tight_layout()
            save_figure(fig_matched, "receiver_mf_time.pdf")

            fig_matched_freq, grid_matched_freq = create_figure(3, 2, figsize=(16, 9))

            ImpulseResponsePlot(
                fig_matched_freq, grid_matched_freq, (0, 0),
                self.matched_filterI.t_rc, self.matched_filterI.g_inverted,
                t_unit="ms",
                colors="darkorange",
                label=r"$g(-t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            ImpulseResponsePlot(
                fig_matched_freq, grid_matched_freq, (0, 1),
                self.matched_filterQ.t_rc, self.matched_filterQ.g_inverted,
                t_unit="ms",
                colors="darkorange",
                label=r"$g(-t)$", 
                xlabel=r"Tempo ($ms$)", 
                ylabel="Amplitude", 
                xlim=(-15, 15),
                amp_norm=True,
            ).plot()

            FrequencyPlot(
                fig_matched_freq, grid_matched_freq, (1, 0),
                fs=self.fs,
                signal=dI_prime,
                fc=self.fc,
                labels=[r"$d_I'(f)$"],
                title=r"Componente $I$ - Fitragem Passa-Baixa",
                xlim=(-10, 10),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_matched_freq, grid_matched_freq, (1, 1),
                fs=self.fs,
                signal=dQ_prime,
                fc=self.fc,
                labels=[r"$d_Q'(f)$"],
                title=r"Componente $Q$ - Fitragem Passa-Baixa",
                xlim=(-10, 10),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_matched_freq, grid_matched_freq, (2, 0),
                fs=self.fs,
                signal=It_prime,
                fc=self.fc,
                labels=[r"$I'(f)$"],
                title=r"Componente $I$ - Fitragem Casada",
                xlim=(-10, 10),
                colors="darkgreen",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            FrequencyPlot(
                fig_matched_freq, grid_matched_freq, (2, 1),
                fs=self.fs,
                signal=Qt_prime,
                fc=self.fc,
                labels=[r"$Q'(f)$"],
                title=r"Componente $Q$ - Fitragem Casada",
                xlim=(-10, 10),
                colors="navy",
                style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot()

            fig_matched_freq.tight_layout()
            save_figure(fig_matched_freq, "receiver_mf_freq.pdf")

        return It_prime, Qt_prime

    def synchronizer(self, It_prime, Qt_prime):
        r"""
        Realiza a sincronização do sinal recebido, retornando o sinal sincronizado.

        Args:
            It_prime (np.ndarray): Sinal $I'(t)$ a ser sincronizado.
            Qt_prime (np.ndarray): Sinal $Q'(t)$ a ser sincronizado.

        Returns:
            delayI (float): Delay do sinal $I'(t)$.
            delayQ (float): Delay do sinal $Q'(t)$.

        Example:
            Tempo: ![pageplot](assets/receiver_sync_time.svg)
            Módulo Correlação: ![pageplot](assets/receiver_sync_corr.svg)
        """

        delayI_min, delayI_max, delayI, corr_vec_I = self.synchronizerI.correlation(It_prime, "I")
        delayQ_min, delayQ_max, delayQ, corr_vec_Q = self.synchronizerQ.correlation(Qt_prime, "Q")

        if self.output_print:
            print("\n ==== SINCRONIZADOR ==== \n")
            print("Delay Q Min  :", delayQ_min)
            print("Delay Q Max  :", delayQ_max)
            print("Delay Q Corr :", delayQ)
            print("Delay I Min  :", delayI_min)
            print("Delay I Max  :", delayI_max)
            print("Delay I Corr :", delayI)

        # Nota: delayI e delayQ devem ser iguais, portanto, delayI é ajustado para ser igual a delayQ
        # O ajuste deve ser feito pois o canal I não suporta sincronização.
        delayI_min, delayI_max, delayI = delayQ_min, delayQ_max, delayQ
        
        if self.output_plot:
            fig_corr, grid_corr = create_figure(1, 1, figsize=(16, 9))
            CorrelationPlot(
                fig_corr, grid_corr, (0, 0),
                corr_vec=corr_vec_Q,  
                fs=self.fs,
                xlim_ms=(0, 300),
                colors="darkblue",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                },
            ).plot()
            fig_corr.tight_layout()
            
            save_figure(fig_corr, "receiver_sync_corr.pdf")
    
            fig_sync, grid_sync = create_figure(2,1, figsize=(16, 9))

            SincronizationPlot(
                fig_sync, grid_sync, (0,0),
                t= np.arange(len(It_prime)) / self.fs,
                signal=It_prime,
                sync_start=delayI_min,
                sync_end=delayI_max,
                max_corr=delayI,
                title=r"Canal $I$",
                labels=[r"$d_I(t)$"],
                colors="darkgreen",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                },
                xlim=(40, 200),
            ).plot()

            SincronizationPlot(
                fig_sync, grid_sync, (1,0),
                t=np.arange(len(Qt_prime)) / self.fs,
                signal=Qt_prime,
                sync_start=delayQ_min,
                sync_end=delayQ_max,
                max_corr=delayQ,
                title=r"Canal $Q$",
                labels=[r"$d_Q(t)$"],
                colors="darkblue",
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                },
                xlim=(40, 200),
            ).plot()

            fig_sync.tight_layout()
            save_figure(fig_sync, "receiver_sync_time.pdf")

        return delayI_max, delayQ_max

    def sampler(self, It_prime, Qt_prime, t):
        r"""
        Realiza a decisão (amostragem e quantização) dos sinais $I'(t)$ e $Q'(t)$, retornando os vetores de simbolos $X'_{NRZ}[n]$ e $Y'_{MAN}[n]$.

        Args:
            It_prime (np.ndarray): Sinal $I'(t)$ a ser amostrado e quantizado.
            Qt_prime (np.ndarray): Sinal $Q'(t)$ a ser amostrado e quantizado.
            t (np.ndarray): Vetor de tempo.

        Returns:
            Xnrz_prime (np.ndarray): Sinal $X'_{NRZ}[n]$ amostrado e quantizado.
            Yman_prime (np.ndarray): Sinal $Y'_{MAN}[n]$ amostrado e quantizado.
        
        Example:
            - Tempo: ![pageplot](assets/receiver_sampler_time.svg)
            - Constelação: ![pageplot](assets/receiver_sampler_const.svg)  
            - Fase: ![pageplot](assets/receiver_sampler_phase.svg)  
        """ 

        s_sampledI = self.samplerI.sample(It_prime)
        t_sampledI = self.samplerI.sample(t)
        Xi_prime = self.samplerI.quantize(s_sampledI)

        s_sampledQ = self.samplerQ.sample(Qt_prime)
        t_sampledQ = self.samplerQ.sample(t)
        Yq_prime = self.samplerQ.quantize(s_sampledQ)

        if self.output_print:
            print("\n ==== DECISOR ==== \n")
            print("X'i:", ' '.join(f"{x:+d}" for x in Xi_prime[:20]),"...")
            print("Y'q:", ' '.join(f"{y:+d}" for y in Yq_prime[:20]),"...")

        if self.output_plot:
            fig_sampler, grid_sampler = create_figure(2, 1, figsize=(16, 9))

            SampledSignalPlot(
                fig_sampler, grid_sampler, (0, 0),
                t,
                It_prime,
                t_sampledI,
                s_sampledI,
                colors='darkgreen',
                label_signal="Sinal original", 
                label_samples="Amostras", 
                xlim=(80, 240), 
                title="Componente $I$ amostrado"
            ).plot()

            SampledSignalPlot(
                fig_sampler, grid_sampler, (1, 0),
                t,
                Qt_prime,
                t_sampledQ,
                s_sampledQ,
                colors='navy',
                label_signal="Sinal original", 
                label_samples="Amostras", 
                xlim=(80, 240), 
                title="Componente $Q$ amostrado"
            ).plot()

            fig_sampler.tight_layout()
            save_figure(fig_sampler, "receiver_sampler_time.pdf")            

            fig_const, grid_const = create_figure(1, 2, figsize=(16, 9))

            ConstellationPlot(
                fig_const, grid_const, (0, 0),
                dI=It_prime[:40000:5],
                dQ=Qt_prime[:40000:5],
                xlim=(-1.4, 1.4),
                ylim=(-1.4, 1.4),
                title="Constelação $IQ$",
                colors=["darkred"],
                style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot() 

            ConstellationPlot(
                fig_const, grid_const, (0, 1),
                dI=s_sampledI,
                dQ=s_sampledQ,
                xlim=(-1.4, 1.4),
                ylim=(-1.4, 1.4),
                title="Constelação $IQ - Amostrado$",
                colors=["darkred"],
                style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
            ).plot() 

            fig_const.tight_layout()
            save_figure(fig_const, "receiver_sampler_const.pdf")

            fig_phase, grid_phase = create_figure(1, 2, figsize=(16, 9))

            PhasePlot(
                fig_phase, grid_phase, (0, 0),
                t=t,
                signals=[It_prime, Qt_prime],
                labels=["Fase $I + jQ$"],
                title="Fase $I + jQ$",
                xlim=(40, 320),
                colors=["darkred"],
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            PhasePlot(
                fig_phase, grid_phase, (0, 1),
                t=t_sampledI,
                signals=[np.array(Xi_prime), np.array(Yq_prime)],
                labels=["Fase $I + jQ$"],
                title="Fase $I + jQ$ - Decidido",
                xlim=(40, 320),
                colors=["darkred"],
                style={
                    "line": {"linewidth": 2, "alpha": 1},
                    "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
                }
            ).plot()

            fig_phase.tight_layout()
            save_figure(fig_phase, "receiver_sampler_phase.pdf")

        return Xi_prime, Yq_prime

    def decode(self, Xnrz_prime, Yman_prime):
        r"""
        Decodifica os vetores de simbolos codificados $X'_{NRZ}[n]$ e $Y'_{MAN}[n]$, retornando os vetores de bits $X'n$ e $Y'n$.

        Args:
            Xnrz_prime (np.ndarray): Sinal $X'_{NRZ}[n]$ quantizado.
            Yman_prime (np.ndarray): Sinal $Y'_{MAN}[n]$ quantizado.

        Returns:
            Xn_prime (np.ndarray): Sinal $X'n$ decodificado.
            Yn_prime (np.ndarray): Sinal $Y'n$ decodificado.
        
        Example:
            - Tempo: ![pageplot](assets/receiver_decoder_time.svg)
        """

        i_quantized = np.array(Xnrz_prime)
        q_quantized = np.array(Yman_prime)
        
        Xn_prime = self.decoderI.decode(i_quantized)
        Yn_prime = self.decoderQ.decode(q_quantized)

        if self.output_print:
            print("\n ==== DECODIFICADOR DE LINHA ==== \n")
            print("X'n:", ''.join(map(str, Xn_prime)))
            print("Y'n:", ''.join(map(str, Yn_prime)))
        
        if self.output_plot:
            fig_decoder, grid_decoder = create_figure(4, 1, figsize=(16, 9))

            SymbolsPlot(
                fig_decoder, grid_decoder, (0, 0),
                symbols_list=[Xnrz_prime],
                samples_per_symbol=1,
                colors=["darkgreen"],
                xlabel="Index de Simbolo",
                xlim=(0, 60),
                ylabel="$X_{NRZ}[n]$", 
                label="$X_{NRZ}[n]$"
            ).plot()

            BitsPlot(
                fig_decoder, grid_decoder, (1, 0),
                bits_list=[Xn_prime],
                sections=[("$X_n$", len(Xn_prime))],
                colors=["darkgreen"],
                xlabel="Index de Bit", 
                ylabel="$X_n$", 
                xlim=(0, 60)
            ).plot()

            SymbolsPlot(
                fig_decoder, grid_decoder, (2, 0),
                symbols_list=[Yman_prime],
                samples_per_symbol=1,
                colors=["navy"],
                xlabel="Index de Simbolo",
                xlim=(0, 60),
                ylabel="$Y_{MAN}[n]$", 
                label="$Y_{MAN}[n]$"
            ).plot()

            BitsPlot(
                fig_decoder, grid_decoder, (3, 0),
                bits_list=[Yn_prime],
                sections=[("$Y_n$", len(Yn_prime))],
                colors=["navy"],
                xlabel="Index de Bit", 
                ylabel="$Y_n$", 
                xlim=(0, 60)
            ).plot()

            fig_decoder.tight_layout()
            save_figure(fig_decoder, "receiver_decoder_time.pdf")
                 
        return Xn_prime, Yn_prime

    def descrambler(self, Xn_prime, Yn_prime):
        r"""
        Desembaralha os vetores de bits $X'n$ e $Y'n$, retornando os vetores de bits $v_{t}^{0'}$ e $v_{t}^{1'}$.

        Args:
            Xn_prime (np.ndarray): Vetor de bits $X'n$ embaralhados.
            Yn_prime (np.ndarray): Vetor de bits $Y'n$ embaralhados.

        Returns:
            vt0 (np.ndarray): Vetor de bits $v_{t}^{0'}$ desembaralhado.
            vt1 (np.ndarray): Vetor de bits $v_{t}^{1'}$ desembaralhado.

        Example:
            - Tempo: ![pageplot](assets/receiver_descrambler_time.svg)
        """

        vt0, vt1 = self.unscrambler.descramble(Xn_prime, Yn_prime)

        if self.output_print:
            print("\n ==== DESEMBARALHADOR ==== \n")
            print("vt0':", ''.join(map(str, vt0)))
            print("vt1':", ''.join(map(str, vt1)))
        
        if self.output_plot:
            fig_descrambler, grid_descrambler = create_figure(4, 1, figsize=(16, 9))

            BitsPlot(
                fig_descrambler, grid_descrambler, (0, 0),
                bits_list=[Xn_prime],
                sections=[("$X_n$", len(Xn_prime))],
                colors=["darkgreen"],
                ylabel="Embaralhado",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_descrambler, grid_descrambler, (1, 0),
                bits_list=[vt0],
                sections=[("$v_t^{0}$", len(vt0))],
                colors=["darkgreen"],
                ylabel="Restaurado", 
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_descrambler, grid_descrambler, (2, 0),
                bits_list=[Yn_prime],
                sections=[("$Y_n$", len(Yn_prime))],
                colors=["navy"],
                ylabel="Embaralhado",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_descrambler, grid_descrambler, (3, 0),
                bits_list=[vt1],
                sections=[("$v_t^{1}$", len(vt1))],
                colors=["navy"],
                ylabel="Restaurado", 
                xlabel="Index de Bit",
                xlim=(0, 60)
            ).plot()

            fig_descrambler.tight_layout()
            save_figure(fig_descrambler, "receiver_descrambler_time.pdf")     

        return vt0, vt1

    def conv_decoder(self, vt0, vt1):
        r"""
        Decodifica os vetores de bits $v_{t}^{0'}$ e $v_{t}^{1'}$, retornando o vetor de bits $u_{t}'$.

        Args:
            vt0 (np.ndarray): Vetor de bits $v_{t}^{0'}$ desembaralhado.
            vt1 (np.ndarray): Vetor de bits $v_{t}^{1'}$ desembaralhado.

        Returns:
            ut (np.ndarray): Vetor de bits $u_{t}'$ decodificado.
        
        Example:
            - Tempo: ![pageplot](assets/receiver_conv_time.svg)
        """

        ut = self.conv_viterbi.decode(vt0, vt1)

        if self.output_print:
            print("\n ==== DECODIFICADOR VITERBI ==== \n")
            print("u't:", ''.join(map(str, ut)))
        
        if self.output_plot:
            fig_conv_decoder, grid_conv_decoder = create_figure(3, 1, figsize=(16, 9))

            BitsPlot(
                fig_conv_decoder, grid_conv_decoder, (0, 0),
                bits_list=[vt0],
                sections=[("$v_t^{0}$", len(vt0))],
                colors=["darkgreen"],
                ylabel="Canal $I$",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_conv_decoder, grid_conv_decoder, (1, 0),
                bits_list=[vt1],
                sections=[("$v_t^{1}$", len(vt1))],
                colors=["navy"],
                ylabel="Canal $Q$",
                xlim=(0, 60)
            ).plot()

            BitsPlot(
                fig_conv_decoder, grid_conv_decoder, (2, 0),
                bits_list=[ut],
                sections=[("$u_t'$", len(ut))],
                colors=["darkred"],
                ylabel="Decodificado", 
                xlabel="Index de Bit",
                xlim=(0, 60)
            ).plot()

            fig_conv_decoder.tight_layout()
            save_figure(fig_conv_decoder, "receiver_conv_time.pdf")     

        return ut

    
    def datagram(self, ut):
        r"""
        Recebe um vetor de bits $u_{t}'$ decodificado e retorna um datagrama no padrão ARGOS-3, ou o vetor de bits $u_{t}'$ se houver erro.

        Args:
            ut (np.ndarray): Vetor de bits $u_{t}'$ decodificado.

        Returns:
            datagram (np.ndarray): Datagrama gerado, ou o vetor de bits $u_{t}'$ se houver erro.
            success (bool): Indica se a operação foi bem-sucedida.

        Example:
            - Tempo: ![pageplot](assets/receiver_datagram_time.svg)
        """
        try:
            datagramRX = Datagram(streambits=ut)

            if self.output_print:
                print("\n ==== DATAGRAMA ==== \n")
                print("\n",datagramRX.parse_datagram())

            if self.output_plot:
                fig_datagram, grid = create_figure(1, 1, figsize=(16, 5))
                BitsPlot(
                    fig_datagram, grid, (0, 0),
                    bits_list=[datagramRX.msglength, 
                               datagramRX.pcdid, 
                               datagramRX.blocks, 
                               datagramRX.tail],
                    sections=[("Message Length", len(datagramRX.msglength)),
                              ("PCD ID", len(datagramRX.pcdid)),
                              ("Dados de App.", len(datagramRX.blocks)),
                              ("Tail", len(datagramRX.tail))],
                    colors=["green", "orange", "red", "blue"],
                    xlabel="Index de Bit",
                    xlim=(0, 60),
                ).plot()
                fig_datagram.tight_layout()
                save_figure(fig_datagram, "receiver_datagram_time.pdf")
            
            return datagramRX, True

        except Exception as e:
            print("Erro ao gerar datagrama:", e)
            return ut, False
    
    def receive(self, s):
        r"""
        Executa o processo de recepção, retornando o resultado da recepção.

        Args:
            s (np.ndarray): Sinal $s(t)$ recebido.

        Returns:
            datagramRX (np.ndarray): Datagrama gerado, ou vetor ut se houver erro.
        """

        t = np.arange(0, len(s)/self.fs, 1/self.fs)

        xI_prime, yQ_prime = self.demodulate(s, t)
        dI_prime, dQ_prime= self.lowpassfilter(xI_prime, yQ_prime, t)
        It_prime, Qt_prime = self.matchedfilter(dI_prime, dQ_prime, t)
        self.delayI, self.delayQ = self.synchronizer(It_prime, Qt_prime)

        # Atualiza o delay do sampler
        self.samplerI.update_sampler(self.delayI, t)
        self.samplerQ.update_sampler(self.delayQ, t)

        Xnrz_prime, Yman_prime = self.sampler(It_prime, Qt_prime, t)
        Xn_prime, Yn_prime = self.decode(Xnrz_prime, Yman_prime)
        vt0, vt1 = self.descrambler(Xn_prime, Yn_prime)
        ut = self.conv_decoder(vt0, vt1)

        datagramRX, success = self.datagram(ut)
        return datagramRX, success 


if __name__ == "__main__":

    fc = np.random.randint(10, 90)*100
    print("SIMULANDO TRANSMISSÃO/RECEPÇÃO COM FC =", fc)

    datagramTX = Datagram(pcdnum=1234, numblocks=1)
    transmitter = Transmitter(fc=fc, output_print=True, output_plot=True)
    t, s = transmitter.transmit(datagramTX)

    print("\n\n ==== CANAL ==== \n")    

    channel = Channel(fs=transmitter.fs, duration=0.335, noise_mode="ebn0", noise_db=20, seed=11)
    channel.add_signal(s, position_factor=1)
    channel.add_noise()

    # Comprimentos (para verificação do canal)
    signal_length = len(s) / transmitter.fs
    print("Comprimento do sinal modulado:", signal_length)

    signalnoise_length = len(channel.channel) / transmitter.fs
    print("Comprimento do sinal recebido:", signalnoise_length)

    noise_first = (signalnoise_length - signal_length)
    print("Comprimento ruido antes do sinal:", noise_first)
    
    receiver = Receiver(fc=fc, output_print=True)
    datagramRX, success = receiver.receive(channel.channel)
        
    if not success:
        bitsTX = datagramTX.streambits 
        bitsRX = datagramRX
        print("Bits TX: ", ''.join(str(b) for b in bitsTX))
        print("Bits RX: ", ''.join(str(b) for b in bitsRX))

        # Calcula a Taxa de Erro de Bit (BER)
        num_errors = sum(1 for tx, rx in zip(bitsTX, bitsRX) if tx != rx)
        ber = num_errors / len(bitsTX)

        print(f"Número de erros: {num_errors}")
        print(f"Taxa de Erro de Bit (BER): {ber:.6f}")




# """
# Implementa um formatador de pulso para transmissão de sinais digitais. 

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
import matplotlib.pyplot as plt

from .preamble import Preamble
from .formatter import Formatter
from .encoder import Encoder
from .plotter import create_figure, save_figure, TimePlot, SincronizationPlot, CorrelationPlot
from .multiplexer import Multiplexer
from .matchedfilter import MatchedFilter

class Synchronizer:
    def __init__(self, fs=128_000, Rb=400, sync_word="2BEEEEBF", channel_encode=("nrz", "man"), sync_window=None):
        r"""
         Inicializa o sincronizador de simbolos para identificar o momento de maior correlação entre o sinal recebido e o sinal de sincronismo.

        Args:
            fs (int): Frequência de amostragem do sinal recebido.
            Rb (int): Taxa de transmissão do sinal recebido.
            sync_word (str): Palavra de sincronismo.
            channel_encode (tuple): Tupla com o tipo de codificação dos canais I e Q respectivamente.

        Example: 
            ![pageplot](assets/example_synchronizer_sync.svg)
        """

        # Validar os valores de channel_encode
        valid_encodings = ["nrz", "man"]
        if channel_encode[0] not in valid_encodings or channel_encode[1] not in valid_encodings:
            raise ValueError("Os tipos de codificação devem ser 'nrz' ou 'manchester'.")
        
        # Parâmetros
        self.fs = fs
        self.Rb = Rb
        self.Tb = 1 / Rb
        self.sps = int(fs / Rb)

        # Parâmetros fixos
        self.alpha = 0.8
        self.span = 6
        self.cI_encoder = "nrz"
        self.cQ_encoder = "nrz"
        self.sync_window = sync_window

        # Codificação I e Q
        self.cI_type = channel_encode[0]
        self.cQ_type = channel_encode[1]

        # Mapeamento das configurações de codificação
        encoding_params = {
            "nrz": {"format": "RRC", "bits_per_symbol": 1, "Rb_multiplier": 1, "matched": "RRC-Inverted"},
            "man": {"format": "Manchester", "bits_per_symbol": 2, "Rb_multiplier": 2, "matched": "Manchester-Inverted"}
        }

        # Parâmetros para o canal I e Q
        cI_params = encoding_params[self.cI_type]
        self.cI_format = cI_params["format"]
        self.cI_bits_per_symbol = cI_params["bits_per_symbol"]
        self.cI_Rb = self.Rb
        self.cI_matched = cI_params["matched"]
        cQ_params = encoding_params[self.cQ_type]
        self.cQ_format = cQ_params["format"]
        self.cQ_bits_per_symbol = cQ_params["bits_per_symbol"]
        self.cQ_Rb = self.Rb
        self.cQ_matched = cQ_params["matched"]


        self.encoder_I = Encoder(method=self.cI_encoder)
        self.encoder_Q = Encoder(method=self.cQ_encoder)
        self.formatterI = Formatter(alpha=self.alpha, fs=self.fs, Rb=self.cI_Rb, span=self.span, type=self.cI_format, channel="I", bits_per_symbol=self.cI_bits_per_symbol)
        self.formatterQ = Formatter(alpha=self.alpha, fs=self.fs, Rb=self.cQ_Rb, span=self.span, type=self.cQ_format, channel="Q", bits_per_symbol=self.cQ_bits_per_symbol)
        self.matched_filter_I = MatchedFilter(alpha=self.alpha, fs=self.fs, Rb=self.cI_Rb, span=self.span, type=self.cI_matched, channel="I", bits_per_symbol=self.cI_bits_per_symbol)
        self.matched_filter_Q = MatchedFilter(alpha=self.alpha, fs=self.fs, Rb=self.cQ_Rb, span=self.span, type=self.cQ_matched, channel="Q", bits_per_symbol=self.cQ_bits_per_symbol)
        self.create_sincronized_word(sync_word)

    def create_sincronized_word(self, sync_word):
        r"""
        Monta os vetores de simbolo $S_I(t)$ e $S_Q(t)$, correspondente a palavra de sincronismo do canal $I$ e $Q$, respectivamente. O comprimento da palavra de sincronismo é dado por $\Delta \tau$, conforme a expressão abaixo.

        $$
        \Delta \tau = L_{sync} \cdot \frac{f_s}{R_b}
        $$

        Sendo: 
            - $\Delta \tau$ é o comprimento da palavra de sincronismo.
            - $L_{sync}$ é o comprimento da palavra de sincronismo de $S_I(t)$ e $S_Q(t)$.
            - $R_b$ é a taxa de bits.
            - $f_s$ é a frequência de amostragem.

        Args:
            sync_word (str): Palavra de sincronismo.

        Example: 
            ![pageplot](assets/example_synchronizer_word.svg)
        """

        self.preamble = Preamble(sync_word)
        self.preamble_sI = self.preamble.preamble_sI
        self.preamble_sQ = self.preamble.preamble_sQ

        self.sincronized_word_I = self.formatterI.apply_format(self.encoder_I.encode(self.preamble_sI), add_prefix=False)
        self.sincronized_word_Q = self.formatterQ.apply_format(self.encoder_Q.encode(self.preamble_sQ), add_prefix=False)

        self.sincronized_word_I = self.matched_filter_I.apply_filter(self.sincronized_word_I)
        self.sincronized_word_Q = self.matched_filter_Q.apply_filter(self.sincronized_word_Q)

    def correlation(self, signal, channel):
        r"""

        Realiza a correlação cruzada entre o sinal recebido $s(t)$ e a palavra de sincronismo $d(t)$, para cada index de tempo $t$.

        $$
        c[k] = \sum_{t=0} s[t] d[t - k]
        $$ 

        Sendo: 
            - $s(t)$ e $d(t)$ são os vetores de simbolos do sinal recebido e da palavra de sincronismo, respectivamente.
            - $k$ é o index de tempo no vetor de correlação cruzada.
            - $c[k]$ é o valor da correlação cruzada para o index $k$.

        Em seguida localiza-se o indice de $c[k]$ com maior valor, resultando em $k_{max}$, este é o indice de amostra com a maior correlação entre o sinal recebido e a palavra de sincronismo, por fim, calcula-se o delay $\tau$. 

        $$
            \tau = \frac{k_{max}}{f_s}
        $$

        Sendo: 
            - $\tau$: Delay entre o sinal recebido e a palavra de sincronismo.
            - $f_s$: Frequência de amostragem do sinal recebido.
            - $k_{max}$: Indice de amostra com a maior correlação entre o sinal recebido e a palavra de sincronismo.

        Args:
            signal (np.ndarray): Sinal recebido.
            channel (str): Canal de recebimento, $I$ ou $Q$.

        Returns:
           delay (tuple): Tupla contendo o delay $\tau$, o delay $\tau_{min}$ e o delay $\tau_{max}$.
        
        Example: 
            ![pageplot](assets/example_synchronizer_corr.svg)
        """
        if channel == "I":
            correlation_vec = np.correlate(signal, self.sincronized_word_I, mode="same")
        elif channel == "Q":
            correlation_vec = np.correlate(signal, self.sincronized_word_Q, mode="same")
        else:
            raise ValueError("Canal inválido. Use 'I' ou 'Q'.")


        # converte janela de segundos para índices
        if self.sync_window is not None:
            t_start, t_end = self.sync_window
            start_idx = int(t_start * self.fs)
            end_idx = int(t_end * self.fs)
            start_idx = max(0, start_idx)
            end_idx = min(len(correlation_vec), end_idx)
        else:
            start_idx, end_idx = 0, len(correlation_vec)        

        local_argmax = correlation_vec[start_idx:end_idx].argmax()
        max_correlation_index = start_idx + local_argmax

        # normaliza o vetor
        correlation_vec = (correlation_vec - correlation_vec.min()) / (correlation_vec.max() - correlation_vec.min())
        
        # calcula o index do início e fim da palavra de sincronismo
        low_index = max_correlation_index - len(self.sincronized_word_I) // 2
        high_index = max_correlation_index + len(self.sincronized_word_I) // 2

        # calcula o delay com base no index: 
        low_delay = low_index / self.fs
        high_delay = high_index / self.fs
        delay = max_correlation_index / self.fs
    
        return low_delay, high_delay, delay, correlation_vec

if __name__ == "__main__":
    
    synchronizer = Synchronizer()

    fig_format, grid_format = create_figure(2,1, figsize=(16, 9))

    TimePlot(
        fig_format, grid_format, (0,0),
        t= np.arange(len(synchronizer.sincronized_word_I)) / synchronizer.formatterI.fs,
        signals=[synchronizer.sincronized_word_I],
        labels=[r"$S_I(t)$"],
        title=r"Canal $I$",
        amp_norm=True,
        colors="darkgreen",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    TimePlot(
        fig_format, grid_format, (1,0),
        t= np.arange(len(synchronizer.sincronized_word_Q)) / synchronizer.formatterQ.fs,
        signals=[synchronizer.sincronized_word_Q],
        labels=[r"$S_Q(t)$"],
        title=r"Canal $Q$",
        amp_norm=True,
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    fig_format.tight_layout()
    save_figure(fig_format, "example_synchronizer_word.pdf")

    # TESTE
    preamble = Preamble()  
    SI = preamble.preamble_sI
    SQ = preamble.preamble_sQ
    X = np.random.randint(0, 2, 60)
    Y = np.random.randint(0, 2, 60)
    
    mux = Multiplexer()
    Xn, Yn = mux.concatenate(SI, SQ, X, Y)

    encoder_nrz = Encoder(method="NRZ")
    encoder_man = Encoder(method="NRZ")
    
    Xnrz = encoder_nrz.encode(Xn)
    Yman = encoder_man.encode(Yn)
    
    formatterI = Formatter(alpha=0.8, fs=128_000, Rb=400, span=12, type="RRC", channel="I", bits_per_symbol=1)
    formatterQ = Formatter(alpha=0.8, fs=128_000, Rb=400, span=12, type="Manchester", channel="Q", bits_per_symbol=2)
    matched_filter_I = MatchedFilter(alpha=0.8, fs=128_000, Rb=400, span=12, type="RRC-Inverted", channel="I", bits_per_symbol=1)
    matched_filter_Q = MatchedFilter(alpha=0.8, fs=128_000, Rb=400, span=12, type="Manchester-Inverted", channel="Q", bits_per_symbol=2)
    
    dI = formatterI.apply_format(Xnrz)
    dQ = formatterQ.apply_format(Yman)
    
    dI = matched_filter_I.apply_filter(dI)
    dQ = matched_filter_Q.apply_filter(dQ)
    
    # Faz a sincronização apenas no canal Q, pois o canal I é apenas uns.
    delayQ_min, delayQ_max, delayQ, corr_vec = synchronizer.correlation(dQ, "Q")
    delayI_min, delayI_max, delayI = delayQ_min, delayQ_max, delayQ

    print("Delay I (ms):", delayI_min)
    print("Delay Q (ms):", delayQ_min)

    fig_sync, grid_sync = create_figure(2,1, figsize=(16, 9))
    
    SincronizationPlot(
        fig_sync, grid_sync, (0,0),
        t= np.arange(len(dI)) / formatterI.fs,
        signal=dI,
        sync_start=delayI_min,
        sync_end=delayI_max,
        max_corr=delayI,
        title=r"Canal $I$",
        labels=[r"$d_I(t)$"],
        colors="darkgreen",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        },
        xlim=(40, 200),
    ).plot()

    SincronizationPlot(
        fig_sync, grid_sync, (1,0),
        t=np.arange(len(dQ)) / formatterQ.fs,
        signal=dQ,
        sync_start=delayQ_min,
        sync_end=delayQ_max,
        max_corr=delayQ,
        title=r"Canal $Q$",
        labels=[r"$d_Q(t)$"],
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        },
        xlim=(40, 200),
    ).plot()

    fig_sync.tight_layout()
    save_figure(fig_sync, "example_synchronizer_sync.pdf")

    fig_corr, grid_corr = create_figure(1, 1, figsize=(16, 9))
    CorrelationPlot(
        fig_corr, grid_corr, (0, 0),
        corr_vec=corr_vec,  
        fs=formatterQ.fs,
        xlim_ms=(40, 200),
        colors="darkblue",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        },
    ).plot()
    fig_corr.tight_layout()
    save_figure(fig_corr, "example_synchronizer_corr.pdf")
    
    
    
        



from importlib import metadata as _metadata

__version__ = _metadata.version("argos3")

# Main Components
from .convolutional import *
from .datagram import *
from .detector import *
from .encoder import *
from .formatter import *
from .lowpassfilter import *
from .matchedfilter import *    
from .modulator import *
from .multiplexer import *
from .preamble import *
from .sampler import * 
from .scrambler import *
from .synchronizer import *

# Channel
from .channel import *
from .noise import *

# Transmitter
from .transmitter import *

# Receiver
from .receiver import *

# Extras
from .bersnr import *
from .data import * 
from .plotter import *



# """
# Implementa um datagrama compatível com o padrão PPT-A3.

# Referência:
#     AS3-SP-516-274-CNES (seção 3.1.4)

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
import json
from .plotter import BitsPlot, create_figure, save_figure

class Datagram: 
    def __init__(self, pcdnum=None, numblocks=None, streambits=None, seed=None, payload=None):
        r"""
        Gera um datagrama no padrão ARGOS-3. O formato do datagrama é ilustrado na figura abaixo.

        ![pageplot](../assets/datagrama.svg)

        Args:
            pcdnum (int): Número identificador da PCD. Necessário para o modo TX.
            numblocks (int): Quantidade de blocos de dados. Necessário para o modo TX.
            streambits (np.ndarray): Sequência de bits do datagrama. Necessário para o modo RX.
            seed (int): Seed do gerador de números aleatórios.
            payload (np.ndarray): Payload do datagrama. 

        Raises:
            ValueError: Se o número de blocos não estiver entre 1 e 8.
            ValueError: Se o número PCD não estiver entre 0 e 1048575 $(2^{20} - 1)$.
            ValueError: Se os parâmetros `pcdnum` e `numblocks` ou `streambits` não forem fornecidos.
            ValueError: Se o payload não for fornecido ou se o comprimento do payload não for o mesmo que o número de blocos.

        Example: 
            ![pageplot](assets/example_datagram_time.svg)

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.2)
        </div>
        """

        # Atributos comuns
        self.streambits = None
        self.blocks_json = None
        
        # O construtor será chamado dependendo de como o datagrama é criado (TX ou RX)
        if pcdnum is not None and numblocks is not None and streambits is None:
            # Construtor TX
            self._init_tx(pcdnum, numblocks, seed, payload)
        elif streambits is not None and pcdnum is None and numblocks is None:
            # Construtor RX
            self._init_rx(streambits)
        else:
            raise ValueError("Você deve fornecer ou (pcdnum e numblocks) ou streambits")
    
    def _init_tx(self, pcdnum, numblocks, seed, payload):
        r"""
        Construtor TX
        """

        if not (1 <= numblocks <= 8):
            raise ValueError("O número de blocos deve estar entre 1 e 8.")
        if not (0 <= pcdnum <= 1048575):  # 2^20 - 1
            raise ValueError("O número PCD deve estar entre 0 e 1048575.")
        if (payload is not None) and (len(payload) != (numblocks -1) * 32 + 24):
            raise ValueError("O payload deve ter o mesmo comprimento que o número de blocos.")
        
        self.pcdnum = pcdnum
        self.numblocks = numblocks
        self.rng = np.random.default_rng(seed)

        # Se não for passado payload, gera os blocos automaticamente
        if payload is not None:
            self.blocks = payload
            # TODO: Relcalcular o tamanho do numblocks.
            # verificar se o numblos está dentro de 1 e 8 
        else:
            self.blocks = self.generate_blocks()

        # Gera os componentes do datagrama
        self.pcdid = self.generate_pcdid()
        self.tail = self.generate_tail()
        self.msglength = self.generate_msglength()

        # A sequência de bits do datagrama
        self.streambits = np.concatenate((self.msglength, self.pcdid, self.blocks, self.tail))

        # Cria a representação JSON do datagrama
        self.blocks_json = self.parse_datagram()

    def _init_rx(self, streambits):
        r"""
        Construtor RX
        """

        self.streambits = streambits
        self.blocks_json = self.parse_datagram()

    def generate_blocks(self):
        r"""
        Gera os blocos de dados simulados (valores aleatórios), com base na quantidade especificada de blocos. 
        
        A quantidade de blocos pode variar de 1 á 8. O primeiro bloco tem comprimento de 24bits, enquanto que todos os demais blocos tem 32bits. Dessa forma, os dados o comprimento dos dados de aplicação são dados pela expressão abaixo.

        $$
        L_{app} = 24 + 32 \cdot (n-1)
        $$

        Sendo: 
            - $L_{app}$: Comprimento do datagrama em bits 
            - $n$: Número de blocos do datagrama, podendo variar de 1 á 8. 

        Returns:
            blocks (np.ndarray): Vetor de bits representando os blocos de dados.

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.2)
        </div>
        """

        length = [24] + [32] * (self.numblocks - 1)
        total_length = sum(length)
        return self.rng.integers(0, 2, size=total_length, dtype=np.uint8)

    def generate_pcdid(self):
        r"""
        Gera o campo $PCD_{ID}$ a partir do número PCD ($PCD_{num}$), Primeiro gera-se a sequência de 20 bits correspondente ao número PCD.

        $$
          PCDnum_{10} \mapsto PCDnum_{2}  
        $$

        Sendo: 
            - $PCDnum_{10}$: Valor decimal do campo $PCD_{num}$, podendo variar de 0 á 1048575 $(2^{20} - 1)$.
            - $PCDnum_{2}$: Sequência de 20 bits correspondente ao valor de $PCD_{num}$.

        Em seguida, é calculado o checksum, $R_{PCD}$, do campo $PCD_{num}$, obtido através da soma dos bits e aplicação da operação módulo 256 ($2^8$).

        $$
        \begin{equation}
        R_{PCD} = \left( \sum_{i=0}^{19} b_i \cdot 2^i \right) \bmod 256
        \end{equation}
        $$

        Sendo: 
            - $R_{PCD}$: Sequência de 8 bits correspondente ao checksum do campo $PCD_{num}$.
            - $i$: Indice do bit do campo $PCD_{num}$.
            - $b$: Valor do bit do campo $PCD_{num}$.

        O campo $PCD_{ID}$ é gerado concatenando os parâmetros gerados, sendo $PCD_{ID} = PCD_{num} \oplus R_{PCD}$.

        Returns:
            pcd_id (np.ndarray): Vetor de bits contendo o PCD ID e o checksum.       

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.2)
        </div>
        """

        bin_str = format(self.pcdnum, '020b')
        pcd_bits = np.array([int(b) for b in bin_str], dtype=np.uint8)

        checksum_val = pcd_bits.sum() % 256
        checksum_bits = np.array([int(b) for b in format(checksum_val, '08b')], dtype=np.uint8)
        return np.concatenate((pcd_bits, checksum_bits))

    def generate_msglength(self):
        r"""
        Gera o valor do comprimento de mensagem $T_{m}$ com base na quantidade de blocos $n$. Primeiro, deve-se calcular a sequência de bits $B_m$. 
        
         $$
           Bm_{10} = (n - 1) \mapsto Bm_{2} 
         $$

        Sendo: 
            - $B_m$: Sequência de três bits correspondendo ao tamanho de mensagem. 
            - $n$: Número de blocos do datagrama, podendo variar de 1 á 8. 

        Em seguida, é calculado o quarto bit $P_m$ (bit de paridade).

        $$
        \begin{equation}
            P_m = 
            \begin{cases}
            1, & \text{se } \left[ \sum_{i=0}^{B_m} b_i = 0 \right]\mod 2  \\
            0, & \text{se } \left[ \sum_{i=0}^{B_m} b_i = 1 \right]\mod 2 
            \end{cases} \text{.}
        \end{equation}
        $$
        
        Sendo: 
            - $P_m$: Bit de paridade.
            - $i$: Indice de bit do campo $B_m$.

        O campo $T_{m}$ é gerado concatenando os parâmetros gerados, sendo $T_{m} = B_{m} \oplus P_{m}$.

        Returns:
           msg_length (np.ndarray): Vetor de 4 bits representando o campo Message Length.

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.2)
        </div>
        """

        n = self.numblocks - 1
        bin_str = format(n, '03b')
        bits = np.array([int(b) for b in bin_str], dtype=np.uint8)
        paridade = bits.sum() % 2
        return np.append(bits, paridade)
    
    def generate_tail(self):
        r"""
        Gera a cauda do datagrama $E_m$, utilizado para limpar o registrador do codificador convolucional.

        $$
        E_m = 7 + [(n - 1) \bmod 3]
        $$

        Sendo: 
            - $E_m$: Comprimento de cauda do datagrama (zeros) adicionada ao final do datagrama. 
            - $n$: Número de blocos do datagrama.

        Returns:
            tail (np.ndarray): Vetor de bits zerados com comprimento variável (7, 8 ou 9 bits).
            
        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.1.4.3)
        </div>
        """

        tail_pad = [7, 8, 9]
        tail_length = tail_pad[(self.numblocks - 1) % 3]
        return np.zeros(tail_length, dtype=np.uint8)

    def parse_datagram(self):
        r"""
        Interpreta a sequência de bits do datagrama, extraindo campos e validando integridade.
        
        Returns:
            str (json): Objeto JSON contendo a representação estruturada do datagrama.
        
        Raises:
            ValueError: Caso haja falha na validação de paridade do comprimento de mensagem $T_m$.
            ValueError: Caso haja falha no checksum do campo $PCD_{ID}$. 
            ValueError: Se a sequência de bits de aplicação não corresponder ao comprimento de $T_m$.

        Example:
            ```python
            >>> datagram = Datagram(streambits=bits)
            >>> print(datagram.parse_datagram())
            {
              "msglength": 2,
              "pcdid": 1234,
              "data": {
                "bloco_1": {
                  "sensor_1": 42,
                  "sensor_2": 147,
                  "sensor_3": 75
                },
                "bloco_2": {
                  "sensor_1": 138,
                  "sensor_2": 7,
                  "sensor_3": 134,
                  "sensor_4": 182
                }
              },
              "tail": 8
            }
            ```
        """

        # extrai o campo message length
        msglength = self.streambits[:4]
        value_bits = msglength[:3]
        paridade_bit = msglength[3]

        # Verifica a integridade do campo
        if paridade_bit != value_bits.sum() % 2:
            raise ValueError("Paridade inválida no campo Message Length.")
        else:
            self.msglength = msglength

        # extrai o campo PCD ID
        pcdid_bits = self.streambits[4:32]
        pcdnum_bits = pcdid_bits[:20]
        checksum_bits = pcdid_bits[20:28]

        # verifica a integridade do campo
        checksum_val = pcdnum_bits.sum() % 256
        if checksum_val != int("".join(map(str, checksum_bits)), 2):
            raise ValueError("Checksum inválido no campo PCD ID.")
        else:
            self.pcdid = pcdid_bits
            self.pcdnum = int("".join(map(str, pcdnum_bits)), 2)            

        
        # extrai o campo dados de aplicação
        self.numblocks = int("".join(map(str, value_bits)), 2) + 1
        self.blocks = self.streambits[32:32 + 24 + (32 * (self.numblocks - 1))]

        # Pega os ultimos bits após dados de app.
        finalbits = self.streambits[32 + 24 + (32 * (self.numblocks - 1)):]

        # verifica a cauda esperada de acordo com numblocks, e extrai ela.
        tail_pad = [7, 8, 9]
        tail_length = tail_pad[(self.numblocks - 1) % 3]
        tail_bits = finalbits[:tail_length]

        # verifica a integridade da cauda, todos os bits tem que ser 0.
        if any(int(b) != 0 for b in tail_bits):
            raise ValueError("Cauda inválida.")
        else:
            self.tail = tail_bits
    
        # cria o objeto JSON
        data = {
            "msglength": self.numblocks,
            "pcdid": self.pcdnum,
            "data": {},
            "tail": tail_length
        }

        # monta o objeto JSON
        index = 0
        for bloco in range(self.numblocks):
            bloco_nome = f"bloco_{bloco+1}"
            data["data"][bloco_nome] = {}
            
            num_sensores = 3 if bloco == 0 else 4
            for sensor in range(num_sensores):
                sensor_nome = f"sensor_{sensor+1}"
                sensor_bits = self.blocks[index:index+8]
                sensor_valor = int("".join(map(str, sensor_bits)), 2)
                data["data"][bloco_nome][sensor_nome] = sensor_valor
                index += 8

        return json.dumps(data, indent=2)

if __name__ == "__main__":
    
    print("\n\nTransmissor:")
    datagram_tx = Datagram(pcdnum=123456, numblocks=2, seed=10)
    print(datagram_tx.parse_datagram())
    print("Stream bits: ", ''.join(str(b) for b in datagram_tx.streambits))

    fig_datagram, grid = create_figure(1, 1, figsize=(16, 5))
    
    BitsPlot(
        fig_datagram, grid, (0, 0),
        bits_list=[datagram_tx.msglength, 
                   datagram_tx.pcdid, 
                   datagram_tx.blocks, 
                   datagram_tx.tail],
        sections=[("Message Length", len(datagram_tx.msglength)),
                  ("PCD ID", len(datagram_tx.pcdid)),
                  ("Dados de App.", len(datagram_tx.blocks)),
                  ("Tail", len(datagram_tx.tail))],
        colors=["green", "orange", "red", "blue"],
        xlabel="Index de Bit"
    ).plot()

    fig_datagram.tight_layout()
    save_figure(fig_datagram, "example_datagram_time.pdf")

    # Receptor
    bits = datagram_tx.streambits

    print("\n\nReceptor: ")
    datagram_rx = Datagram(streambits=bits)
    print(datagram_rx.parse_datagram())
    print("Stream bits: ", ''.join(str(b) for b in datagram_rx.streambits))


    # Teste com payload:

    # Gera um vetor com 24 uns
    payload = np.ones(24, dtype=np.uint8)

    datagram_tx = Datagram(pcdnum=123456, numblocks=1, payload=payload, seed=10)
    print(datagram_tx.parse_datagram())
    print("Stream bits: ", ''.join(str(b) for b in datagram_tx.streambits))

    bits = datagram_tx.streambits

    print("\n\nReceptor: ")
    datagram_rx = Datagram(streambits=bits)
    print(datagram_rx.parse_datagram())
    print("Stream bits: ", ''.join(str(b) for b in datagram_rx.streambits))



# """
# Import e Export de dados para arquivos numpy. 

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import os
import numpy as np

class ExportData:
    r"""
    Instância um objeto `ExportData`, utilizado para salvar vetores em arquivos binários `.npy` ou texto `.txt`.

    Args:
        vector (Union[np.ndarray, List[np.ndarray]]): Um único vetor ou lista de vetores para salvar.
        filename (str): Nome do arquivo de saída.
        path (str): Caminho do diretório de saída.
    """
    def __init__(self, vector, filename, path="../../out"):
        # Converte um único vetor para uma lista com um elemento
        self.vectors = [vector] if isinstance(vector, np.ndarray) else list(vector)
        self.filename = filename
        self.path = path

    def save(self, binary=True):
        r"""
        Salva os resultados em arquivo binário `.npy` ou em texto `.txt`.
        
        Args:
            binary (bool): Se `True`, salva em formato binário `.npy`.
        """
        script_dir = os.path.dirname(os.path.abspath(__file__))
        basepath = os.path.normpath(os.path.join(script_dir, self.path, self.filename))
        os.makedirs(os.path.dirname(basepath), exist_ok=True)

        if binary:
            # Salva em formato binário do NumPy
            # Se houver apenas um vetor, salva como array 1D, senão como array 2D
            data = self.vectors[0] if len(self.vectors) == 1 else np.array(self.vectors)
            np.save(f"{basepath}.npy", data)
        else:
            # Salva em texto (menos eficiente, mas legível)
            with open(f"{basepath}.txt", "w") as f:
                for i, vec in enumerate(self.vectors):
                    if i > 0:
                        f.write("\n--- Vector {} ---\n".format(i+1))
                    f.write(" ".join(map(str, vec)))

class ImportData:
    r"""
    Instância um objeto `ImportData`, utilizado para carregar vetores de arquivos binários `.npy` ou texto `.txt`.

    Args:
        filename (str): Nome do arquivo (sem extensão).
        path (str): Caminho do diretório de entrada.
    """
    def __init__(self, filename, path="../../out"):
        self.filename = filename
        self.path = path

    def load(self, mode="npy", dtype=np.float64):
        r"""
        Carrega o vetor salvo.

        Args:
            mode (str): Formato do arquivo: `npy` para arquivos binários, `txt` para arquivos de texto.
            dtype (np.dtype): Tipo de codificação utilizada, necessário para `npy`.

        Returns:
            data (np.ndarray): Vetor carregado.
        """
        script_dir = os.path.dirname(os.path.abspath(__file__))
        basepath = os.path.normpath(os.path.join(script_dir, self.path, self.filename))

        if mode == "npy":
            return np.load(f"{basepath}.npy")

        elif mode == "bin":
            return np.fromfile(f"{basepath}.bin", dtype=dtype)

        elif mode == "txt":
            with open(f"{basepath}.txt", "r") as f:
                data = list(map(float, f.read().split()))
            return np.array(data, dtype=dtype)

        else:
            raise ValueError(f"Formato '{mode}' não suportado. Use 'npy', 'bin' ou 'txt'.")



# """
# Codificação e decodificação convolucional segundo o padrão CCSDS 131.1-G-2, utilizado no sistema PTT-A3.

# Referência:
#     AS3-SP-516-274-CNES (seção 3.1.4.4)

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
import komm 
from .plotter import create_figure, save_figure, BitsPlot

class EncoderConvolutional: 
    def __init__(self, G=np.array([[0b1111001, 0b1011011]])):
        r"""
        Inicializa o codificador convolucional, com base em uma tupla de polinômios geradores $G$ que determinam a estrutura do codificador.

        $$
        \begin{equation}
            \begin{split}
                G_0 &= 121_{10} \quad \mapsto \quad G_0 = [1, 1, 1, 1, 0, 0, 1] \\
                G_1 &= 91_{10} \quad \mapsto \quad G_1 = [1, 0, 1, 1, 0, 1, 1]
            \end{split}
        \end{equation}
        $$

        O codificador convolucional, considerando $G_0$ e $G_1$ pode ser representado pelo diagrama de blocos abaixo.

        ![pageplot](../assets/cod_convolucional.svg)
        
        Args:
            G (np.ndarray): Tupla de polinômios geradores $G$.

        Example: 
            ![pageplot](assets/example_conv_time.svg)

        <div class="referencia">
          <b>Referência:</b>
          <p>AS3-SP-516-274-CNES (seção 3.1.4.4)</p>
          <p>CCSDS 131.1-G-2</p>
        </div>
        """
        self.G = G
        self.G0 = int(G[0][0])
        self.G1 = int(G[0][1])
        self.K = max(self.G0.bit_length(), self.G1.bit_length())
        self.g0_taps = self.calc_taps(self.G0)
        self.g1_taps = self.calc_taps(self.G1)
        self.shift_register = np.zeros(self.K, dtype=int)
        self.komm = komm.ConvolutionalCode(G)

    def calc_taps(self, poly):
        r"""
        Calcula os índices dos bits ativos ($'1'$), ou taps, do polinômio gerador $G_n$.

        Args:
            poly (int): Polinômio gerador $G_n$ em formato binário. 

        Returns:
            taps (int): Lista com os índices dos taps ativos.
        """
        bin_str = f"{poly:0{self.K}b}"
        taps = [i for i, b in enumerate(bin_str) if b == '1']
        return taps

    def calc_free_distance(self):
        r"""
        Calcula a distância livre $d_{free}$ do código convolucional, definida como a menor distância de Hamming entre quaisquer duas sequências de saída distintas.

        Returns:
            dist (int): Distância livre $d_{free}$ do codificador convolucional organizado com $G$.
        """
        return self.komm.free_distance()

    def encode(self, ut):
        r"""
        Codifica uma sequência binária de entrada $u_t$, retornando as sequências de saida $v_t^{(0)}$ e $v_t^{(1)}$. O processo de codificação pode ser representado pela expressão abaixo.

        $$
        \begin{equation}
        \begin{bmatrix} v_t^{(0)} & v_t^{(1)} \end{bmatrix}
        =
        \begin{bmatrix}
        u_{(t)} & u_{(t-1)} & u_{(t-2)} & u_{(t-3)} & u_{(t-4)} & u_{(t-5)} & u_{(t-6)}
        \end{bmatrix}
        \cdot
        \begin{bmatrix} G_{0} & G_{1} \end{bmatrix}^{T}
        \end{equation}
        $$

        Sendo: 
            - $v_t^{(0)}$ e $v_t^{(1)}$: Canais de saída do codificador.
            - $u_t$: Vetor de bits de entrada.
            - $G_{0}$ e $G_{1}$: Polinômios geradores do codificador.

        Args:
            ut (np.ndarray): Vetor de bits $u_t$ de entrada a serem codificados.

        Returns:
                tuple (np.ndarray, np.ndarray): Tupla com os dois canais de saída $v_t^{(0)}$ e $v_t^{(1)}$.
        """
        ut = np.array(ut, dtype=int)
        vt0 = []
        vt1 = []

        for bit in ut:
            self.shift_register = np.insert(self.shift_register, 0, bit)[:self.K]
            out0 = np.sum(self.shift_register[self.g0_taps]) % 2
            out1 = np.sum(self.shift_register[self.g1_taps]) % 2
            vt0.append(out0)
            vt1.append(out1)

        return np.array(vt0, dtype=int), np.array(vt1, dtype=int)


class DecoderViterbi:
    def __init__(self, G=np.array([[0b1111001, 0b1011011]])):
        r"""
        Inicializa o decodificador convolucional (algoritmo Viterbi), com base em uma tupla de polinômios geradores $G$ que determinam a estrutura do decodificador.

        $$
        \begin{equation}
            \begin{split}
                G_0 &= 121_{10} \quad \mapsto \quad G_0 = [1, 1, 1, 1, 0, 0, 1] \\
                G_1 &= 91_{10} \quad \mapsto \quad G_1 = [1, 0, 1, 1, 0, 1, 1]
            \end{split}
        \end{equation}
        $$

        Args:
            G (np.ndarray): Tupla de polinômios geradores $G$.

        <div class="referencia">
          <b>Referência:</b>
          <p>https://rwnobrega.page/apontamentos/codigos-convolucionais/</p>
          <p>AS3-SP-516-274-CNES (seção 3.1.4.4)</p>
        </div>
        """
        
        self.G = G
        self.G0 = int(G[0][0])
        self.G1 = int(G[0][1])
        self.K = max(self.G0.bit_length(), self.G1.bit_length())
        self.num_states = 2**(self.K - 1)
        self.trellis = self.build_trellis()

    def build_trellis(self):
        r"""
        Constroi a trelica do decodificador Viterbi.

        Returns:
            trellis (dict): Trelica do decodificador Viterbi.
        """
        trellis = {}
        for state in range(self.num_states):
            trellis[state] = {}
            for bit in [0, 1]:
                # reconstruir o shift register (bit atual + estado anterior)
                sr = [bit] + [int(b) for b in format(state, f'0{self.K - 1}b')]
                out0 = sum([sr[i] for i in range(self.K) if (self.G0 >> (self.K - 1 - i)) & 1]) % 2
                out1 = sum([sr[i] for i in range(self.K) if (self.G1 >> (self.K - 1 - i)) & 1]) % 2
                out = [out0, out1]

                # remove último bit para próximo estado
                next_state = int(''.join(str(b) for b in sr[:-1]), 2)  
                trellis[state][bit] = (next_state, out)
        return trellis

    def decode(self, vt0, vt1):
        r"""
        Decodifica os bits de entrada $v_t^{(0)}$ e $v_t^{(1)}$, retornando os bits decodificados $u_t$.

        Args:
            vt0 (np.ndarray): Bits de entrada do canal I.
            vt1 (np.ndarray): Bits de entrada do canal Q.

        Returns:
            ut_hat (np.ndarray): Bits decodificados.
        """
        vt0 = np.array(vt0, dtype=int)
        vt1 = np.array(vt1, dtype=int)
        T = len(vt0)

        # Inicializar métricas
        path_metrics = np.full((T + 1, self.num_states), np.inf)
        path_metrics[0][0] = 0
        prev_state = np.full((T + 1, self.num_states), -1, dtype=int)
        prev_input = np.full((T + 1, self.num_states), -1, dtype=int)

        # Viterbi
        for t in range(T):
            for state in range(self.num_states):
                if path_metrics[t, state] < np.inf:
                    for bit in [0, 1]:
                        next_state, expected_out = self.trellis[state][bit]
                        dist = (expected_out[0] != vt0[t]) + (expected_out[1] != vt1[t])
                        metric = path_metrics[t, state] + dist
                        if metric < path_metrics[t + 1, next_state]:
                            path_metrics[t + 1, next_state] = metric
                            prev_state[t + 1, next_state] = state
                            prev_input[t + 1, next_state] = bit

        # Traceback
        state = np.argmin(path_metrics[T])
        ut_hat = []
        for t in range(T, 0, -1):
            bit = prev_input[t, state]
            ut_hat.append(bit)
            state = prev_state[t, state]

        return np.array(ut_hat[::-1], dtype=int)


if __name__ == "__main__":

    encoder = EncoderConvolutional()
    print("Distância livre:", encoder.calc_free_distance())
    print("G0:  ", format(encoder.G0, 'b'), " |  Taps: ", ''.join(str(b) for b in encoder.g0_taps))
    print("G1:  ", format(encoder.G1, 'b'), " |  Taps: ", ''.join(str(b) for b in encoder.g1_taps))

    ut = np.random.randint(0, 2, 40)
    vt0, vt1 = encoder.encode(ut)
    print("ut:  ", ''.join(str(b) for b in ut))
    print("vt0: ", ''.join(str(b) for b in vt0))
    print("vt1: ", ''.join(str(b) for b in vt1))
    
    fig_conv, grid_conv = create_figure(3, 1, figsize=(16, 9))
    
    BitsPlot(
        fig_conv, grid_conv, (0, 0),
        bits_list=[ut],
        sections=[("$u_t$", len(ut))],
        colors=["darkred"],
        ylabel="$u_t$"
    ).plot()

    BitsPlot(
        fig_conv, grid_conv, (1, 0),
        bits_list=[vt0],
        sections=[("$v_t^{(0)}$", len(vt0))],
        colors=["darkgreen"],
        ylabel="$v_t^{(0)}$"
    ).plot()

    BitsPlot(
        fig_conv, grid_conv, (2, 0),
        bits_list=[vt1],
        sections=[("$v_t^{(1)}$", len(vt1))],
        colors=["navy"],
        xlabel="Index de Bit", 
        ylabel="$v_t^{(1)}$"
    ).plot()

    fig_conv.tight_layout()
    save_figure(fig_conv, "example_conv_time.pdf")

    decoder = DecoderViterbi()
    ut_prime = decoder.decode(vt0, vt1)

    print("ut': ", ''.join(str(b) for b in ut_prime))
    print("ut = ut': ", np.array_equal(ut, ut_prime))
    



# """
# Implementa uma classe de filtro passa baixa para remover a componente de frequência alta do sinal recebido.

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from scipy.signal import butter, filtfilt, lfilter
from .plotter import create_figure, save_figure, ImpulseResponsePlot, TimePlot, PoleZeroPlot, FrequencyResponsePlot

class LPF:
    def __init__(self, cut_off=600, order=6, fs=128_000, type="butter"):
        r"""
        Inicializa um filtro passa-baixa com base em uma frequência de corte $f_{cut}$ e uma ordem $N$.

        Args:
            cut_off (float): Frequência de corte $f_{cut}$ do filtro.
            order (int): Ordem $N$ do filtro.
            fs (int, opcional): Frequência de amostragem $f_s$. 
            type (str, opcional): Tipo de filtro. Padrão é "butter".
        
        Raises:
            ValueError: Se o tipo de filtro for inválido.

        Example: 
            ![pageplot](assets/example_lpf_signals.svg) 
        """

        self.cut_off = cut_off
        self.order = order
        self.fs = fs

        type = type.lower()
        if type != "butter":
            raise ValueError("Tipo de filtro inválido. Use 'butter'.")

        # Coeficientes b (numerador) e a (denominador) do filtro
        self.b, self.a = self.butterworth_filter()
        self.impulse_response, self.t_impulse = self.calc_impulse_response()

    def butterworth_filter(self, fNyquist=0.5):
        r"""
        Calcula os coeficientes do filtro Butterworth utilizando a biblioteca `scipy.signal`. A função de transferência contínua $H(s)$ de um filtro Butterworth é dada pela expressão abaixo.

        $$
        H(s) = \frac{1}{1 + \left(\frac{s}{2 \pi f_{cut}}\right)^{2n}}
        $$

        Sendo:
            - $s$: Variável complexa no domínio de Laplace.
            - $2 \pi f_{cut}$: Frequência angular de corte do filtro.
            - $n$: Ordem do filtro.

        Args: 
            fNyquist (float): Fator de Nyquist. Padrão é 0.5 * fs.

        Returns:
            tuple: Coeficientes $b$ e $a$ correspondentes à função de transferência do filtro Butterworth.

        Example:
            ![pageplot](assets/example_lpf_pz.svg)
        """
        b, a = butter(self.order, self.cut_off / (fNyquist * self.fs), btype='low')
        return b, a

    def calc_impulse_response(self, impulse_len=1024):
        r"""
        Para obter a resposta ao impulso no dominio do tempo, um impulso unitário é aplicado como entrada. Para um filtro Butterworth, o calculo é dado pela expressão abaixo. 

        $$
        h(t) = \mathcal{L}^{-1}\left\{H(f)\right\}
        $$

        Sendo:
            - $h(t)$: Resposta ao impulso do filtro.
            - $H(f)$: Função de transferência do filtro.
            - $\mathcal{L}^{-1}$: Transformada de Laplace inversa.

        Args: 
            impulse_len (int): Comprimento do vetor de impulso.

        Returns:
            impulse_response (tuple[np.ndarray, np.ndarray]): Resposta ao impulso e vetor de tempo.
        
        Example: 
            ![pageplot](assets/example_lpf_impulse.svg)
        """
        # Impulso unitário
        impulse_input = np.zeros(impulse_len)
        impulse_input[0] = 1

        # Resposta ao impulso
        impulse_response = lfilter(self.b, self.a, impulse_input)
        t_impulse = np.arange(impulse_len) / self.fs
        return impulse_response, t_impulse

    def apply_filter(self, signal):
        r"""
        Aplica o filtro passa-baixa com resposta ao impulso $h(t)$ ao sinal de entrada $s(t)$, utilizando a função `scipy.signal.filtfilt`. O processo de filtragem é dado pela expressão abaixo. 

        $$
            x(t) = s(t) \ast h(t)
        $$

        Sendo: 
            - $x(t)$: Sinal filtrado.
            - $s(t)$: Sinal de entrada.
            - $h(t)$: Resposta ao impulso do filtro.

        Args:
            signal (np.ndarray): Sinal de entrada $s(t)$.

        Returns:
            signal_filtered (np.ndarray): Sinal filtrado $x(t)$.
        """
        signal_filtered = filtfilt(self.b, self.a, signal)

        # normalização
        # signal_filtered *= np.sqrt(2)

        return signal_filtered


if __name__ == "__main__":
    
    fs = 128_000
    t = np.arange(10000) / fs

    # create two cossine signals with different frequencies
    f1 = 1000
    f2 = 4000
    signal = np.cos(2 * np.pi * f1 * t) + np.cos(2 * np.pi * f2 * t)

    filtro = LPF(cut_off=1500, order=6, fs=fs, type="butter")
    signal_filtered = filtro.apply_filter(signal)

    fig_impulse, grid_impulse = create_figure(1, 1, figsize=(16, 5))

    ImpulseResponsePlot(
        fig_impulse, grid_impulse, (0, 0),
        filtro.t_impulse, filtro.impulse_response,
        t_unit="ms",
        colors="darkorange",
        label=r"$h(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", xlim=(0, 5), 
        amp_norm=True
    ).plot()

    fig_impulse.tight_layout()
    save_figure(fig_impulse, "example_lpf_impulse.pdf")

    fig_signal, grid_signal = create_figure(2, 2, figsize=(16, 9))

    ImpulseResponsePlot(
        fig_signal, grid_signal, (0, slice(0, 2)),
        filtro.t_impulse, filtro.impulse_response,
        t_unit="ms",
        colors="darkorange",
        label=r"$h(t)$", 
        xlabel=r"Tempo ($ms$)", 
        ylabel="Amplitude", 
        xlim=(0, 5), 
        amp_norm=True
    ).plot()
    
    TimePlot(
        fig_signal, grid_signal, (1, 0),
        t, 
        signal,
        labels=[r"$x(t)$"],
        title="Sinal original",
        xlim=(0, 8),
        amp_norm=True,
        colors="navy"
    ).plot()

    TimePlot(
        fig_signal, grid_signal, (1, 1),
        t, 
        signal_filtered,
        labels=[r"$x'(t)$"],
        title="Sinal filtrado",
        xlim=(0, 8),
        amp_norm=True,
        colors="darkred"
    ).plot()

    fig_signal.tight_layout()
    save_figure(fig_signal, "example_lpf_signals.pdf")

    fig_pz, grid_pz = create_figure(1, 1, figsize=(10,10))
    PoleZeroPlot(
            fig_pz, grid_pz, (0,0), 
            filtro.b, filtro.a,
            colors="darkblue",
            title="Polos e Zeros",
        ).plot()
    save_figure(fig_pz, "example_lpf_pz.pdf")

    freq_response, grid_freq_response = create_figure(1, 1, figsize=(16,6))
    FrequencyResponsePlot(
            freq_response, grid_freq_response, (0,0), 
            filtro.b, filtro.a, 
            fs=filtro.fs, 
            f_cut=filtro.cut_off, 
            xlim=(0, 3*filtro.cut_off)
        ).plot()
    save_figure(freq_response, "example_lpf_freq_response.pdf")
    


# """
# Implementação de um modulador IQ para transmissão de sinais digitais.

# Autor: Arthur Cadore
# Data: 28-07-2025
# """

import numpy as np
from .formatter import Formatter
from .encoder import Encoder
from .plotter import PhasePlot, create_figure, save_figure, TimePlot, FrequencyPlot, ConstellationPlot 
from scipy.signal import hilbert

class Modulator:
    def __init__(self, fc=None, fs=128_000):
        r"""
        Inicializa um modulador QPSK no padrão ARGOS-3. O modulador pode ser representado pelo diagrama de blocos apresentado abaixo.

        ![pageplot](../assets/modulador.svg)

        Args:
            fc (float): Frequência da portadora.
            fs (int): Frequência de amostragem.

        Raises:
            ValueError: Se a frequência de amostragem não for maior que o dobro da frequência da portadora. (Teorema de Nyquist)
       
        Example: 
            ![pageplot](assets/transmitter_modulator_time.svg)

        <div class="referencia">
        <b>Referência:</b><br>
        AS3-SP-516-274-CNES (seção 3.2.5.3)
        </div>
        """
        if fc is None or fc <= 0:
            raise ValueError("A frequência da portadora deve ser maior que zero.")
        
        if fs <= fc*2:
            raise ValueError("A frequência de amostragem deve ser maior que o dobro da frequência da portadora.")
        
        self.fc = fc
        self.fs = fs

    def modulate(self, i_signal, q_signal):
        r"""
        Modula em QPSK os sinais $d_I$ e $d_Q$ com uma portadora $f_c$, resultando no sinal modulado $s(t)$. O processo de modulação é dado pela expressão abaixo.

        $$
            s(t) = d_I(t) \cdot \cos(2\pi f_c t) - d_Q(t) \cdot \sin(2\pi f_c t)
        $$

        Sendo: 
            - $s(t)$: Sinal modulado.
            - $d_I(t)$ e $d_Q(t)$: Sinais formatados correspondentes aos canais $I$ e $Q$.
            - $f_c$: Frequência da portadora.
            - $t$: Vetor de tempo.

        Args:
            i_signal (np.ndarray): Sinal $d_I$ correspondente ao canal $I$ a ser modulado.
            q_signal (np.ndarray): Sinal $d_Q$ correspondente ao canal $Q$ a ser modulado.

        Returns:
            modulated_signal (np.ndarray): Sinal modulado $s(t)$ resultante.
            t (np.ndarray): Vetor de tempo $t$ correspondente ao sinal modulado.

        Raises:
            ValueError: Se os sinais I e Q não tiverem o mesmo tamanho.
        """
        n = len(i_signal)
        if len(q_signal) != n:
            raise ValueError("i_signal e q_signal devem ter o mesmo tamanho.")
        
        t = np.arange(n) / self.fs
        carrier_cos = np.cos(2 * np.pi * self.fc * t)
        carrier_sin = np.sin(2 * np.pi * self.fc * t)
        
        modulated_signal = (i_signal * carrier_cos - q_signal * carrier_sin)

        return t, modulated_signal
    
    def demodulate(self, modulated_signal, carrier_length=0.07, carrier_delay=0):
        r"""
        Demodula o sinal modulado em QPSK, com sincronização de fase usando os primeiros `carrier_length` segundos de portadora.

        Args:
            modulated_signal (np.ndarray): Sinal modulado $s(t)$ a ser demodulado.
            carrier_length (float): O comprimento da portadora para sincronização da fase (em segundos).
            carrier_delay (float): O delay da portadora para sincronização da fase (em segundos).

        Returns:
            i_signal (np.ndarray): Sinal $x_I'(t)$ recuperado.
            q_signal (np.ndarray): Sinal $y_Q'(t)$ recuperado.
        """
        # Verifica se o sinal não está vazio
        n = len(modulated_signal)
        if n == 0:
            raise ValueError("O sinal modulado não pode estar vazio.")

        # Vetor de tempo
        t = np.arange(n) / self.fs

        # Pega um subvetor do sinal modulado para estimar a fase
        carrier_signal = modulated_signal[int(carrier_delay * self.fs):(int(carrier_delay * self.fs) + int(carrier_length * self.fs))]

        # Aplicando a Transformada de Hilbert para obter a fase instantânea
        analytic_signal = hilbert(carrier_signal)
        original_phase = np.angle(analytic_signal)

        # Estima a fase e corrige o sinal
        phase_estimate = np.mean(original_phase)
        modulated_signal_corrected = modulated_signal * np.exp(-1j * phase_estimate)

        # Calcula os componentes de demodulação
        carrier_cos = 2 * np.cos(2 * np.pi * self.fc * t)
        carrier_sin = 2 * np.sin(2 * np.pi * self.fc * t)

        # Demodula o sinal
        i_signal = modulated_signal_corrected * carrier_cos
        q_signal = -modulated_signal_corrected * carrier_sin

        # Correção de polaridade
        if np.mean(i_signal) < 0:
            i_signal = -i_signal
            q_signal = -q_signal

        return np.real(i_signal), np.real(q_signal), phase_estimate, original_phase


if __name__ == "__main__":

    fs = 128_000
    fc = 4000
    Rb = 400
    alpha = 0.8
    span = 8

    Xnrz = np.random.randint(0, 2, 200)
    Yman = np.random.randint(0, 2, 200)

    encoder_nrz = Encoder(method="NRZ")
    encoder_man = Encoder(method="NRZ")

    Xnrz = encoder_nrz.encode(Xnrz)
    Yman = encoder_man.encode(Yman)

    print("Xnrz:", ''.join(str(b) for b in Xnrz[:20]))
    print("Yman:", ''.join(str(b) for b in Yman[:20]))

    formatterI = Formatter(alpha=alpha, fs=fs, Rb=Rb, type="RRC", span=span, channel="I", bits_per_symbol=1)
    formatterQ = Formatter(alpha=alpha, fs=fs, Rb=Rb, type="Manchester", span=span, channel="Q", bits_per_symbol=2)
    
    dI = formatterI.apply_format(Xnrz)
    dQ = formatterQ.apply_format(Yman)

    print("dI:", ''.join(str(b) for b in dI[:5]))
    print("dQ:", ''.join(str(b) for b in dQ[:5]))
    
    modulator = Modulator(fc=fc, fs=fs)
    t, s = modulator.modulate(dI, dQ)
    print("s:", ''.join(str(b) for b in s[:5]))

    # PLOT 1 - Tempo
    fig_time, grid = create_figure(2, 1, figsize=(16, 9))
    TimePlot(
        fig_time, grid, (0, 0),
        t=t,
        signals=[dI, dQ],
        labels=["$dI(t)$", "$dQ(t)$"],
        title="Sinal $IQ$ - Formatados RRC",
        xlim=(40, 140),
        amp_norm=True,
        colors=["darkgreen", "navy"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    TimePlot(
        fig_time, grid, (1, 0),
        t=t,
        signals=[s],
        labels=["$s(t)$"],
        title="Sinal Modulado $IQ$",
        xlim=(40, 140),
        amp_norm=True,
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    fig_time.tight_layout()
    save_figure(fig_time, "example_modulator_time.pdf")

    # PLOT 2 - Frequência
    fig_freq, grid = create_figure(2, 2, figsize=(16, 9))
    FrequencyPlot(
        fig_freq, grid, (0, 0),
        fs=fs,
        signal=dI,
        fc=fc,
        labels=["$D_I(f)$"],
        title="Componente I",
        xlim=(-1.5, 1.5),
        colors="navy",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    FrequencyPlot(
        fig_freq, grid, (0, 1),
        fs=fs,
        signal=dQ,
        fc=fc,
        labels=["$D_Q(f)$"],
        title="Componente Q",
        xlim=(-1.5, 1.5),
        colors="darkgreen",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    FrequencyPlot(
        fig_freq, grid, (1, slice(0, 2)),
        fs=fs,
        signal=s,
        fc=fc,
        labels=["$S(f)$"],
        title="Sinal Modulado $IQ$",
        xlim=(-10, 10),
        colors="darkred",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    fig_freq.tight_layout()
    save_figure(fig_freq, "example_modulator_freq.pdf")

    # PLOT 3 - Constelação
    fig_const, grid = create_figure(1, 2, figsize=(16, 8))
    PhasePlot(
        fig_const, grid, (0, 0),
        t=t,
        signals=[dI, dQ],
        labels=["Fase $I + jQ$"],
        title="Fase $I + jQ$",
        xlim=(40, 140),
        colors=["darkred"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()

    ConstellationPlot(
        fig_const, grid, (0, 1),
        dI=dI[:20000:5],
        dQ=dQ[:20000:5],
        title="Constelação $IQ$",
        xlim=(-1.2, 1.2),
        ylim=(-1.2, 1.2),
        colors=["darkred"],
        style={"line": {"linewidth": 2, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    fig_const.tight_layout()
    save_figure(fig_const, "example_modulator_constellation.pdf")

    # Plot 4 - Portadora pura e sinal modulado
    fig_portadora, grid = create_figure(1, 2, figsize=(16, 8))
    FrequencyPlot(
        fig_portadora, grid, (0, 0),
        fs=fs,
        signal=s[0:(int(round(0.082 * fs)))],
        fc=fc,
        labels=["$S(f)$"],
        title="Portadora Pura - $0$ a $80$ms",
        xlim=(-10, 10),
        colors="darkred",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    FrequencyPlot(
        fig_portadora, grid, (0, 1),
        fs=fs,
        signal=s[(int(round(0.082 * fs))):],
        fc=fc,
        labels=["$S(f)$"],
        title="Sinal Modulado - $80$ms em diante",
        xlim=(-10, 10),
        colors="darkred",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    fig_portadora.tight_layout()
    save_figure(fig_portadora, "example_modulator_portadora.pdf")
    
    # Demodulação
    i_signal, q_signal = modulator.demodulate(s)
    print("i_signal:", ''.join(str(b) for b in i_signal[:5]))
    print("q_signal:", ''.join(str(b) for b in q_signal[:5]))

    # PLOT 1 - Tempo
    fig_time, grid = create_figure(2, 1, figsize=(16, 9))
    TimePlot(
        fig_time, grid, (0, 0),
        t=t,
        signals=[i_signal, q_signal],
        labels=["$xI'(t)$", "$yQ'(t)$"],
        title="Componentes $IQ$ - Demoduladas",
        xlim=(40, 140),
        amp_norm=True,
        colors=["darkgreen", "navy"],
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    TimePlot(
        fig_time, grid, (1, 0),
        t=t,
        signals=[s],
        labels=["$s(t)$"],
        title="Sinal Modulado $IQ$",
        xlim=(40, 140),
        amp_norm=True,
        colors="darkred",
        style={
            "line": {"linewidth": 2, "alpha": 1},
            "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}
        }
    ).plot()
    
    fig_time.tight_layout()
    save_figure(fig_time, "example_demodulator_time.pdf")
    

    # PLOT 2 - Frequência
    fig_freq, grid = create_figure(3, 1, figsize=(16, 9))
    FrequencyPlot(
        fig_freq, grid, (0, 0),
        fs=fs,
        signal=s,
        fc=fc,
        labels=["$S(f)$"],
        title="Sinal Modulado $IQ$",
        xlim=(-10, 10),
        colors="darkred",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    
    FrequencyPlot(
        fig_freq, grid, (1, 0),
        fs=fs,
        signal=i_signal,
        fc=fc,
        labels=["$X_I'(f)$"],
        title="Componente $I$ - Demodulado",
        xlim=(-10, 10),
        colors="darkgreen",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()

    FrequencyPlot(
        fig_freq, grid, (2, 0),
        fs=fs,
        signal=q_signal,
        fc=fc,
        labels=["$Y_Q'(f)$"],
        title="Componente $Q$ - Demodulado",
        xlim=(-10, 10),
        colors="navy",
        style={"line": {"linewidth": 1, "alpha": 1}, "grid": {"color": "gray", "linestyle": "--", "linewidth": 0.5}}
    ).plot()
    

    fig_freq.tight_layout()
    save_figure(fig_freq, "example_demodulator_freq.pdf")